"""Open a TCX file in Garmin BaseCamp."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

from idbutils import OpenWithApp


class OpenWithBaseCamp(OpenWithApp):
    """Class that opens a file with an application regardsless of platform."""

    @classmethod
    def _open_on_darwin(cls, filename):
        """Open a file with MacOS application."""
        # cls.open_on_darwin('Garmin BaseCamp.app', filename)
        cls.open_on_darwin_with_applescript('Garmin BaseCamp', f'open POSIX file "{filename}"')



"""Version information for the application."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import idbutils.version as uv

from .version_info import python_required, python_tested, version_info, prerelease


version_string = uv.to_string(version_info, prerelease)


def format_version(program):
    """Print version information for the script."""
    return uv.format(program, version_string)


def log_version(program):
    """Print version information for the script."""
    uv.log(program, version_string)


def python_version_check(program):
    """Validate the Python version requirements."""
    uv.python_version_check(program, python_required, python_tested)



"""Objects for importing Garmin activity data from Garmin Connect downloads and FIT files."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import sys
import logging
from tqdm import tqdm
import dateutil.parser
import traceback

import fitfile
from idbutils import FileProcessor, JsonFileProcessor

from .garmin_connect_enums import Event, get_summary_sport, get_details_sport
from .garmindb import GarminDb, Device, File, ActivitiesDb, Activities, ActivityRecords, ActivityLaps, StepsActivities, PaddleActivities, CycleActivities
from .tcx import Tcx
from .fit_data import FitData


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class GarminActivitiesFitData(FitData):
    """Class for importing Garmin activity data from FIT files."""

    def __init__(self, input_dir, latest, measurement_system, debug):
        """
        Return an instance of GarminActivitiesFitData.

        Parameters:
        ----------
        db_params (dict): configuration data for accessing the database
        input_dir (string): directory (full path) to check for data files
        latest (Boolean): check for latest files only
        measurement_system (enum): which measurement system to use when importing the files
        debug (Boolean): enable debug logging

        """
        super().__init__(input_dir, debug, latest, False, [fitfile.FileType.activity], measurement_system)


class GarminTcxData():
    """Class for importing Garmin activity data from TCX files."""

    def __init__(self, input_dir, latest, measurement_system, debug):
        """
        Return an instance of GarminTcxData.

        Parameters:
        ----------
        db_params (dict): configuration data for accessing the database
        input_dir (string): directory (full path) to check for data files
        latest (Boolean): check for latest files only
        measurement_system (enum): which measurement system to use when importing the files
        debug (Boolean): enable debug logging

        """
        logger.info("Processing activities tcx data")
        self.measurement_system = measurement_system
        self.debug = debug
        if input_dir:
            self.file_names = FileProcessor.dir_to_files(input_dir, Tcx.filename_regex, latest)

    def file_count(self):
        """Return the number of files that will be propcessed."""
        return len(self.file_names)

    def __process_record(self, tcx, activity_id, record_number, point):
        root_logger.debug("Processing record: %r (%d)", point, record_number)
        if not ActivityRecords.s_exists(self.garmin_act_db_session, {'activity_id' : activity_id, 'record' : record_number}):
            record = {
                'activity_id'                       : activity_id,
                'record'                            : record_number,
                'timestamp'                         : tcx.get_point_time(point),
                'hr'                                : tcx.get_point_hr(point),
                'altitude'                          : tcx.get_point_altitude(point).meters_or_feet(measurement_system=self.measurement_system),
                'speed'                             : tcx.get_point_speed(point).kph_or_mph(measurement_system=self.measurement_system)
            }
            loc = tcx.get_point_loc(point)
            if loc is not None:
                record.update({'position_lat': loc.lat_deg, 'position_long': loc.long_deg})
            self.garmin_act_db_session.add(ActivityRecords(**record))

    def __process_lap(self, tcx, activity_id, lap_number, lap):
        root_logger.info("Processing lap: %d", lap_number)
        for record_number, point in enumerate(tcx.get_lap_points(lap)):
            self.__process_record(tcx, activity_id, record_number, point)
        if not ActivityLaps.s_exists(self.garmin_act_db_session, {'activity_id' : activity_id, 'lap' : lap_number}):
            lap_data = {
                'activity_id'                       : activity_id,
                'lap'                               : lap_number,
                'start_time'                        : tcx.get_lap_start(lap),
                'stop_time'                         : tcx.get_lap_end(lap),
                'elapsed_time'                      : tcx.get_lap_duration(lap),
                'distance'                          : tcx.get_lap_distance(lap).meters_or_feet(measurement_system=self.measurement_system),
                'calories'                          : tcx.get_lap_calories(lap)
            }
            start_loc = tcx.get_lap_start_loc(lap)
            if start_loc is not None:
                lap_data.update({'start_lat': start_loc.lat_deg, 'start_long': start_loc.long_deg})
            end_loc = tcx.get_lap_end_loc(lap)
            if end_loc is not None:
                lap_data.update({'stop_lat': end_loc.lat_deg, 'stop_long': end_loc.long_deg})
            root_logger.info("Inserting lap: %r (%d): %r", lap, lap_number, lap_data)
            self.garmin_act_db_session.add(ActivityLaps(**lap_data))

    def __process_file(self, file_name):
        tcx = Tcx()
        tcx.read(file_name)
        start_time = tcx.start_time
        (manufacturer, product) = tcx.get_manufacturer_and_product()
        serial_number = tcx.serial_number
        device = {
            'serial_number'     : serial_number,
            'timestamp'         : start_time,
            'manufacturer'      : manufacturer,
            'product'           : product,
            'hardware_version'  : None,
        }
        Device.s_insert_or_update(self.garmin_db_session, device, ignore_none=True)
        root_logger.info("Processing file: %s for manufacturer %s product %s device %s", file_name, manufacturer, product, serial_number)
        (file_id, file_name) = File.name_and_id_from_path(file_name)
        file = {
            'id'            : file_id,
            'name'          : file_name,
            'type'          : File.FileType.tcx,
            'serial_number' : serial_number,
        }
        File.s_insert_or_update(self.garmin_db_session, file)
        activity = {
            'activity_id'               : file_id,
            'name'                      : file_id,
            'start_time'                : start_time,
            'stop_time'                 : tcx.end_time,
            'laps'                      : tcx.lap_count,
            'sport'                     : tcx.sport,
            'calories'                  : tcx.calories,
            'distance'                  : tcx.distance.kms_or_miles(self.measurement_system),
            'avg_hr'                    : tcx.hr_avg,
            'max_hr'                    : tcx.hr_max,
            'max_cadence'               : tcx.cadence_max,
            'avg_cadence'               : tcx.cadence_avg,
            'ascent'                    : tcx.ascent.meters_or_feet(self.measurement_system),
            'descent'                   : tcx.descent.meters_or_feet(self.measurement_system)
        }
        start_loc = tcx.start_loc
        if start_loc is not None:
            activity.update({'start_lat': start_loc.lat_deg, 'start_long': start_loc.long_deg})
        end_loc = tcx.end_loc
        if end_loc is not None:
            activity.update({'stop_lat': end_loc.lat_deg, 'stop_long': end_loc.long_deg})
        Activities.s_insert_or_update(self.garmin_act_db_session, activity, ignore_none=True, ignore_zero=True)
        for lap_number, lap in enumerate(tcx.laps):
            self.__process_lap(tcx, file_id, lap_number, lap)

    def process_files(self, db_params):
        """Import data from TCX files into the database."""
        garmin_db = GarminDb(db_params, self.debug - 1)
        garmin_act_db = ActivitiesDb(db_params, self.debug - 1)
        with garmin_db.managed_session() as self.garmin_db_session, garmin_act_db.managed_session() as self.garmin_act_db_session:
            for file_name in tqdm(self.file_names, unit='files'):
                try:
                    self.__process_file(file_name)
                except Exception as e:
                    logger.error('Failed to processes TCX file %s: %s', file_name, e)
                    root_logger.error('Failed to processes TCX file %s: %s', file_name, traceback.format_exc())


class GarminJsonActivityData(JsonFileProcessor):
    """Base class for importing Garmin activity data from JSON formatted Garmin Connect details downloads."""

    def __init__(self, db_params, file_regex, input_dir, latest, measurement_system, debug):
        """
        Return an instance of GarminJsonDetailsData.

        Parameters:
        ----------
        db_params (dict): configuration data for accessing the database
        input_dir (string): directory (full path) to check for data files
        latest (Boolean): check for latest files only
        measurement_system (enum): which measurement system to use when importing the files
        debug (Boolean): enable debug logging

        """
        super().__init__(file_regex, input_dir=input_dir, latest=latest, debug=debug)
        self.measurement_system = measurement_system
        self.garmin_act_db = ActivitiesDb(db_params, self.debug - 1)
        self.conversions = {}

    def _process_common(self, json_data):
        distance = self._get_field_obj(json_data, 'distance', fitfile.Distance.from_meters)
        ascent = self._get_field_obj(json_data, 'elevationGain', fitfile.Distance.from_meters)
        descent = self._get_field_obj(json_data, 'elevationLoss', fitfile.Distance.from_meters)
        avg_speed = self._get_field_obj(json_data, 'averageSpeed', fitfile.Speed.from_mps)
        max_speed = self._get_field_obj(json_data, 'maxSpeed', fitfile.Speed.from_mps)
        max_temperature = self._get_field_obj(json_data, 'maxTemperature', fitfile.Temperature.from_celsius)
        min_temperature = self._get_field_obj(json_data, 'minTemperature', fitfile.Temperature.from_celsius)
        avg_temperature = self._get_field_obj(json_data, 'averageTemperature', fitfile.Temperature.from_celsius)
        start_time = dateutil.parser.parse(self._get_field(json_data, 'startTimeLocal'), ignoretz=True)
        elapsed_time = fitfile.conversions.secs_to_dt_time(self._get_field(json_data, 'elapsedDuration', int))
        return {
            'start_time'                : start_time,
            'stop_time'                 : start_time + fitfile.conversions.time_to_timedelta(elapsed_time),
            'elapsed_time'              : elapsed_time,
            'moving_time'               : fitfile.conversions.secs_to_dt_time(self._get_field(json_data, 'movingDuration', int)),
            'start_lat'                 : self._get_field(json_data, 'startLatitude', float),
            'start_long'                : self._get_field(json_data, 'startLongitude', float),
            'stop_lat'                  : self._get_field(json_data, 'endLatitude', float),
            'stop_long'                 : self._get_field(json_data, 'endLongitude', float),
            'distance'                  : distance.kms_or_miles(self.measurement_system),
            'laps'                      : self._get_field(json_data, 'lapCount'),
            'avg_hr'                    : self._get_field(json_data, 'averageHR', float),
            'max_hr'                    : self._get_field(json_data, 'maxHR', float),
            'calories'                  : self._get_field(json_data, 'calories', float),
            'avg_speed'                 : avg_speed.kph_or_mph(self.measurement_system) if avg_speed is not None else None,
            'max_speed'                 : max_speed.kph_or_mph(self.measurement_system) if max_speed is not None else None,
            'ascent'                    : ascent.meters_or_feet(self.measurement_system) if ascent is not None else None,
            'descent'                   : descent.meters_or_feet(self.measurement_system) if descent is not None else None,
            'max_temperature'           : max_temperature.c_or_f(self.measurement_system) if max_temperature is not None else None,
            'min_temperature'           : min_temperature.c_or_f(self.measurement_system) if min_temperature is not None else None,
            'avg_temperature'           : avg_temperature.c_or_f(self.measurement_system) if avg_temperature is not None else None,
            'training_effect'           : self._get_field(json_data, 'aerobicTrainingEffect', float),
            'anaerobic_training_effect' : self._get_field(json_data, 'anaerobicTrainingEffect', float),
            'max_rr'                    : self._get_field(json_data, 'maxRespirationRate', float),
            'avg_rr'                    : self._get_field(json_data, 'avgRespirationRate', float),
        }

    def process(self):
        """Import data from files into the database."""
        with self.garmin_act_db.managed_session() as self.garmin_act_db_session:
            self._process_files()


class GarminJsonSummaryData(GarminJsonActivityData):
    """Class for importing Garmin activity data from JSON formatted Garmin Connect summary downloads."""

    def __init__(self, db_params, input_dir, latest, measurement_system, debug):
        """
        Return an instance of GarminTcxData.

        Parameters:
        ----------
        db_params (dict): configuration data for accessing the database
        input_dir (string): directory (full path) to check for data files
        latest (Boolean): check for latest files only
        measurement_system (enum): which measurement system to use when importing the files
        debug (Boolean): enable debug logging

        """
        logger.info("Processing %s activities summary data from %s", 'latest' if latest else 'all', input_dir)
        super().__init__(db_params, r'activity_\d*\.json', input_dir, latest, measurement_system, debug)

    def _process_steps_activity(self, activity_id, activity_summary):
        root_logger.debug("process_steps_activity for %s", activity_id)
        avg_vertical_oscillation = self._get_field_obj(activity_summary, 'avgVerticalOscillation', fitfile.Distance.from_meters)
        avg_step_length = self._get_field_obj(activity_summary, 'avgStrideLength', fitfile.Distance.from_meters)
        run = {
            'activity_id'               : activity_id,
            'steps'                     : self._get_field(activity_summary, 'steps', float),
            'avg_steps_per_min'         : self._get_field(activity_summary, 'averageRunningCadenceInStepsPerMinute', float),
            'max_steps_per_min'         : self._get_field(activity_summary, 'maxRunningCadenceInStepsPerMinute', float),
            'avg_step_length'           : avg_step_length.meters_or_feet(self.measurement_system),
            'avg_gct_balance'           : self._get_field(activity_summary, 'avgGroundContactBalance', float),
            'avg_vertical_oscillation'  : avg_vertical_oscillation.meters_or_feet(self.measurement_system),
            'avg_ground_contact_time'   : fitfile.conversions.ms_to_dt_time(self._get_field(activity_summary, 'avgGroundContactTime', float)),
            'vo2_max'                   : self._get_field(activity_summary, 'vO2MaxValue', float),
        }
        StepsActivities.s_insert_or_update(self.garmin_act_db_session, run, ignore_none=True)

    def _process_inline_skating(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("inline_skating for %s: %r", activity_id, activity_summary)

    def _process_snowshoeing(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("snow_shoe for %s: %r", activity_id, activity_summary)

    def _process_strength_training(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("strength_training for %s: %r", activity_id, activity_summary)

    def _process_stand_up_paddleboarding(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("stand_up_paddleboarding for %s: %r", activity_id, activity_summary)

    def _process_resort_skiing_snowboarding(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("resort_skiing_snowboarding for %s: %r", activity_id, activity_summary)

    def _process_running(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("process_running for %s", activity_id)
        self._process_steps_activity(activity_id, activity_summary)
    #
    # def _process_treadmill_running(self, sub_sport, activity_id, activity_summary):
    #     root_logger.debug("process_treadmill_running for %s", activity_id)
    #     self._process_steps_activity(activity_id, activity_summary)

    def _process_walking(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("process_walking for %s", activity_id)
        self._process_steps_activity(activity_id, activity_summary)

    def _process_hiking(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("process_hiking for %s", activity_id)
        self._process_steps_activity(activity_id, activity_summary)

    def _process_paddling(self, sub_sport, activity_id, activity_summary):
        activity = {
            'activity_id'               : activity_id,
            'avg_cadence'               : self._get_field(activity_summary, 'avgStrokeCadence', float),
            'max_cadence'               : self._get_field(activity_summary, 'maxStrokeCadence', float),
        }
        Activities.s_insert_or_update(self.garmin_act_db_session, activity, ignore_none=True)
        avg_stroke_distance = fitfile.Distance.from_meters(self._get_field(activity_summary, 'avgStrokeDistance', float))
        paddle = {
            'activity_id'               : activity_id,
            'strokes'                   : self._get_field(activity_summary, 'strokes', float),
            'avg_stroke_distance'       : avg_stroke_distance.meters_or_feet(self.measurement_system),
        }
        PaddleActivities.s_insert_or_update(self.garmin_act_db_session, paddle, ignore_none=True)

    def _process_cycling(self, sub_sport, activity_id, activity_summary):
        activity = {
            'activity_id'               : activity_id,
            'avg_cadence'               : self._get_field(activity_summary, 'averageBikingCadenceInRevPerMinute', float),
            'max_cadence'               : self._get_field(activity_summary, 'maxBikingCadenceInRevPerMinute', float),
        }
        Activities.s_insert_or_update(self.garmin_act_db_session, activity, ignore_none=True)
        ride = {
            'activity_id'               : activity_id,
            'strokes'                   : self._get_field(activity_summary, 'strokes', float),
            'vo2_max'                   : self._get_field(activity_summary, 'vO2MaxValue', float),
        }
        CycleActivities.s_insert_or_update(self.garmin_act_db_session, ride, ignore_none=True)

    def _process_mountain_biking(self, sub_sport, activity_id, activity_summary):
        return self._process_cycling(sub_sport, activity_id, activity_summary)

    def _process_fitness_equipment(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("process_fitness_equipment (%s) for %s", sub_sport, activity_id)
        self._call_process_func(sub_sport.name, None, activity_id, activity_summary)

    def _process_json(self, json_data):
        activity_id = json_data['activityId']
        event = Event.from_json(json_data)
        sport, sub_sport = get_summary_sport(json_data)
        activity = {
            'activity_id'               : activity_id,
            'name'                      : json_data.get('activityName'),
            'description'               : self._get_field(json_data, 'description'),
            'type'                      : event.name,
            'sport'                     : sport.name,
            'sub_sport'                 : sub_sport.name,
            'laps'                      : self._get_field(json_data, 'lapCount'),
        }
        activity.update(self._process_common(json_data))
        Activities.s_insert_or_update(self.garmin_act_db_session, activity, ignore_none=True)
        self._call_process_func(sport.name, sub_sport, activity_id, json_data)
        return 1


class GarminJsonDetailsData(GarminJsonActivityData):
    """Class for importing Garmin activity data from JSON formatted Garmin Connect details downloads."""

    def __init__(self, db_params, input_dir, latest, measurement_system, debug):
        """
        Return an instance of GarminJsonDetailsData.

        Parameters:
        ----------
        db_params (dict): configuration data for accessing the database
        input_dir (string): directory (full path) to check for data files
        latest (Boolean): check for latest files only
        measurement_system (enum): which measurement system to use when importing the files
        debug (Boolean): enable debug logging

        """
        logger.info("Processing activities detail data")
        super().__init__(db_params, r'activity_details_\d*\.json', input_dir, latest, measurement_system, debug)

    def _process_steps_activity(self, sub_sport, activity_id, json_data):
        summary_dto = json_data['summaryDTO']
        avg_speed = fitfile.conversions.mps_to_mph(summary_dto.get('averageSpeed'))
        avg_moving_speed = fitfile.conversions.mps_to_mph(summary_dto.get('averageMovingSpeed'))
        max_speed = fitfile.conversions.mps_to_mph(summary_dto.get('maxSpeed'))
        run = {
            'activity_id'       : activity_id,
            'avg_pace'          : fitfile.conversions.perhour_speed_to_pace(avg_speed),
            'avg_moving_pace'   : fitfile.conversions.perhour_speed_to_pace(avg_moving_speed),
            'max_pace'          : fitfile.conversions.perhour_speed_to_pace(max_speed),
        }
        root_logger.debug("steps_activity for %d: %r", activity_id, run)
        StepsActivities.s_insert_or_update(self.garmin_act_db_session, run, ignore_none=True)

    def _process_cycling(self, sub_sport, activity_id, json_data):
        root_logger.debug("cycling (%s) for %d: %r", sub_sport, activity_id, json_data)

    def _process_elliptical(self, sub_sport, activity_id, json_data):
        root_logger.debug("elliptical for %d: %r", activity_id, json_data)

    def _process_hiking(self, sub_sport, activity_id, json_data):
        root_logger.debug("hiking for %d: %r", activity_id, json_data)
        self._process_steps_activity(sub_sport, activity_id, json_data)

    def _process_inline_skating(self, sub_sport, activity_id, json_data):
        root_logger.debug("inline_skating for %d: %r", activity_id, json_data)

    def _process_paddling(self, sub_sport, activity_id, json_data):
        root_logger.debug("paddling for %d: %r", activity_id, json_data)
    #
    # def _process_mountain_biking(self, sub_sport, activity_id, json_data):
    #     root_logger.debug("mountain_biking for %d: %r", activity_id, json_data)

    def _process_resort_skiing_snowboarding(self, sub_sport, activity_id, json_data):
        root_logger.debug("resort_skiing_snowboarding for %d: %r", activity_id, json_data)

    def _process_snowshoeing(self, sub_sport, activity_id, json_data):
        root_logger.debug("snow_shoe for %d: %r", activity_id, json_data)

    def _process_strength_training(self, sub_sport, activity_id, json_data):
        root_logger.debug("strength_training for %d: %r", activity_id, json_data)

    def _process_stand_up_paddleboarding(self, sub_sport, activity_id, json_data):
        root_logger.debug("stand_up_paddleboarding for %d: %r", activity_id, json_data)
    #
    # def _process_treadmill_running(self, sub_sport, activity_id, json_data):
    #     root_logger.debug("treadmill_running for %d: %r", activity_id, json_data)
    #     self._process_steps_activity(sub_sport, activity_id, json_data)

    def _process_running(self, sub_sport, activity_id, json_data):
        root_logger.debug("running (%s) for %d: %r", sub_sport, activity_id, json_data)
        self._process_steps_activity(sub_sport, activity_id, json_data)

    def _process_walking(self, sub_sport, activity_id, json_data):
        root_logger.debug("walking (%s) for %d: %r", sub_sport, activity_id, json_data)
        self._process_steps_activity(sub_sport, activity_id, json_data)

    def _process_fitness_equipment(self, sub_sport, activity_id, json_data):
        root_logger.debug("fitness_equipment (%s) for %d: %r", sub_sport, activity_id, json_data)
        self._call_process_func(sub_sport.name, None, activity_id, json_data)

    def _process_json(self, json_data):
        activity_id = json_data['activityId']
        metadata_dto = json_data['metadataDTO']
        summary_dto = json_data['summaryDTO']
        sport, sub_sport = get_details_sport(json_data)
        activity = {
            'activity_id'   : activity_id,
            'course_id'     : self._get_field(metadata_dto, 'associatedCourseId', int)
        }
        activity.update(self._process_common(summary_dto))
        Activities.s_insert_or_update(self.garmin_act_db_session, activity, ignore_none=True)
        self._call_process_func(sport.name, sub_sport, activity_id, json_data)
        return 1



"""Class that manages Garmin Connect download config."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import sys
import platform
import subprocess
import datetime

from idbutils import JsonConfig

from .statistics import Statistics
from .config_manager import ConfigManager


class GarminConnectConfigManager(JsonConfig):
    """Class that manages Garmin Connect downloads."""

    def __init__(self):
        """Return a new GarminConnectConfigManager instance."""
        self.enabled_statistics = None
        try:
            super().__init__(ConfigManager.get_config_file())
        except Exception as e:
            print(str(e))
            print("Missing config: copy GarminConnectConfig.json.example to GarminConnectConfig.json and edit GarminConnectConfig.json to "
                  "add your Garmin Connect username and password.")
            sys.exit(-1)

    def __get_node_value(self, node, leaf):
        node = self.config.get(node)
        if node is not None:
            return node.get(leaf)

    def __get_node_value_default(self, node, leaf, default):
        node = self.config.get(node)
        if node is not None:
            return node.get(leaf, default)
        return default

    def get_secure_password(self):
        """Return the Garmin Connect password from secure storage. On MacOS that si the KeyChain."""
        system = platform.system()
        if system == 'Darwin':
            password = subprocess.check_output(["security", "find-internet-password", "-s", "sso.garmin.com", "-w"])
            if password:
                return password.rstrip()

    def get_user(self):
        """Return the Garmin Connect username."""
        return self.__get_node_value('credentials', 'user')

    def get_password(self):
        """Return the Garmin Connect password."""
        password = self.__get_node_value('credentials', 'password')
        if not password:
            password = self.get_secure_password()
        return password

    def latest_activity_count(self):
        """Return the number of activities to download when getting the latest."""
        return self.__get_node_value('data', 'download_latest_activities')

    def all_activity_count(self):
        """Return the number of activities to download when getting all activities."""
        return self.__get_node_value('data', 'download_all_activities')

    def stat_start_date(self, stat_type):
        """Return a tuple containing the start date and the number of days to fetch stats from."""
        date = self.__get_node_value('data', stat_type + '_start_date')
        days = (datetime.datetime.now().date() - date).days
        return (date, days)

    def device_mount_dir(self):
        """Return the directory where the Garmin USB device is mounted."""
        return self.__get_node_value('copy', 'mount_dir')

    def download_days_overlap(self):
        """Return the number of days to overlap previously downloaded data when downloading."""
        return self.__get_node_value('data', 'download_days_overlap')

    def course_views(self, type):
        """Return a list of course ids to create views for for the given activitiy type."""
        return self.__get_node_value('course_views', type)

    def is_stat_enabled(self, statistic):
        """Return whether a particular statistic is enabled or not."""
        return statistic in self.enabled_stats()

    def enabled_stats(self):
        """Return all enabled statistics as a list of string names."""
        if not self.enabled_statistics:
            json_enabled_stats_dict = self.config.get('enabled_stats', {stat_name: True for stat_name in list(Statistics)})
            self.enabled_statistics = [Statistics.from_string(stat_name) for stat_name, stat_enabled in json_enabled_stats_dict.items() if stat_enabled]
        return self.enabled_statistics



"""Functions for managing the application config."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import os
import logging
import tempfile

from idbutils import DbParams

from .config import Config


logger = logging.getLogger(__name__)


class ConfigManager(Config):
    """Provides accessors to the base class config."""

    temp_dir = tempfile.mkdtemp()
    dev_db = None
    dev_tables = {}

    @classmethod
    def get_db_type(cls):
        """Return the type (SQLite, MySQL, etc) of database that is configured."""
        return cls.db['type']

    @classmethod
    def get_db_user(cls):
        """Return the configured username of the database."""
        return cls.db['user']

    @classmethod
    def get_db_password(cls):
        """Return the configured password of the database."""
        return cls.db['password']

    @classmethod
    def get_db_host(cls):
        """Return the configured hostname of the database."""
        return cls.db['host']

    @classmethod
    def _create_dir_if_needed(cls, dir):
        if not os.path.exists(dir):
            os.makedirs(dir)
        return dir

    @classmethod
    def get_config_dir(cls):
        """Return the configured directory of where the configuation files will be stored."""
        config = cls.directories['config_dir']
        if cls.directories['relative_to_home']:
            homedir = os.path.expanduser('~')
            return homedir + os.sep + config
        return config

    @classmethod
    def get_config_file(cls):
        """Return the path to the configuation file."""
        return cls.get_config_dir() + os.sep + 'GarminConnectConfig.json'

    @classmethod
    def get_base_dir(cls, test_dir=False):
        """Return the configured directory of where the data files will be stored."""
        base = cls.directories['base_dir']
        if test_dir:
            return cls.temp_dir + os.sep + base
        if cls.directories['relative_to_home']:
            homedir = os.path.expanduser('~')
            return homedir + os.sep + base
        return base

    @classmethod
    def get_fit_files_dir(cls, test_dir=False):
        """Return the configured directory of where the FIT files will be stored."""
        return cls.get_base_dir(test_dir) + os.sep + cls.directories['fit_file_dir']

    @classmethod
    def get_or_create_fit_files_dir(cls, test_dir=False):
        """Return the configured directory of where the FIT files will be stored creating it if needed."""
        return cls._create_dir_if_needed(cls.get_fit_files_dir(test_dir))

    @classmethod
    def get_monitoring_base_dir(cls, test_dir=False):
        """Return the configured directory of where the all monitoring files will be stored."""
        return cls.get_fit_files_dir(test_dir) + os.sep + cls.directories['monitoring_file_dir']

    @classmethod
    def get_or_create_monitoring_base_dir(cls, test_dir=False):
        """Return the configured directory of where the monitoring files will be stored creating it if needed."""
        return cls._create_dir_if_needed(cls.get_monitoring_base_dir(test_dir))

    @classmethod
    def get_monitoring_dir(cls, year, test_dir=False):
        """Return the configured directory of where the new monitoring files will be stored."""
        return cls.get_monitoring_base_dir(test_dir) + os.sep + str(year)

    @classmethod
    def get_or_create_monitoring_dir(cls, year, test_dir=False):
        """Return the configured directory of where the monitoring files will be stored creating it if needed."""
        return cls._create_dir_if_needed(cls.get_monitoring_dir(year, test_dir))

    @classmethod
    def get_activities_dir(cls, test_dir=False):
        """Return the configured directory of where the activities files will be stored."""
        return cls.get_fit_files_dir(test_dir) + os.sep + cls.directories['activities_file_dir']

    @classmethod
    def get_or_create_activities_dir(cls, test_dir=False):
        """Return the configured directory of where the activities files will be stored creating it if needed."""
        return cls._create_dir_if_needed(cls.get_activities_dir(test_dir))

    @classmethod
    def get_sleep_dir(cls, test_dir=False):
        """Return the configured directory of where the sleep files will be stored."""
        return cls.get_base_dir(test_dir) + os.sep + cls.directories['sleep_files_dir']

    @classmethod
    def get_or_create_sleep_dir(cls, test_dir=False):
        """Return the configured directory of where the sleep files will be stored creating it if needed."""
        return cls._create_dir_if_needed(cls.get_sleep_dir(test_dir))

    @classmethod
    def get_weight_dir(cls, test_dir=False):
        """Return the configured directory of where the weight files will be stored."""
        return cls.get_base_dir() + os.sep + cls.directories['weight_files_dir']

    @classmethod
    def get_or_create_weight_dir(cls, test_dir=False):
        """Return the configured directory of where the weight files will be stored creating it if needed."""
        return cls._create_dir_if_needed(cls.get_weight_dir(test_dir))

    @classmethod
    def get_rhr_dir(cls, test_dir=False):
        """Return the configured directory of where the resting heart rate files will be stored."""
        return cls.get_base_dir(test_dir) + os.sep + cls.directories['rhr_files_dir']

    @classmethod
    def get_or_create_rhr_dir(cls, test_dir=False):
        """Return the configured directory of where the resting heart rate files will be stored creating it if needed."""
        return cls._create_dir_if_needed(cls.get_rhr_dir(test_dir))

    @classmethod
    def get_fitbit_dir(cls, test_dir=False):
        """Return the configured directory of where the FitBit will be stored."""
        return cls.get_base_dir(test_dir) + os.sep + cls.directories['fitbit_file_dir']

    @classmethod
    def get_or_create_fitbit_dir(cls, test_dir=False):
        """Return the configured directory of where the FitBit files will be stored creating it if needed."""
        return cls._create_dir_if_needed(cls.get_fitbit_dir(test_dir))

    @classmethod
    def get_mshealth_dir(cls, test_dir=False):
        """Return the configured directory of where the Microsoft Health will be stored."""
        return cls.get_base_dir(test_dir) + os.sep + cls.directories['mshealth_file_dir']

    @classmethod
    def get_or_create_mshealth_dir(cls, test_dir=False):
        """Return the configured directory where the Microsoft Health files will be stored creating it if needed."""
        return cls._create_dir_if_needed(cls.get_mshealth_dir(test_dir))

    @classmethod
    def get_plugins_dir(cls):
        """Return the configured directory where the plugin files are located."""
        return cls.get_base_dir() + os.sep + cls.directories['plugins_dir']

    @classmethod
    def get_or_create_plugins_dir(cls):
        """Return the configured directory where the plugin files are located creating it if needed."""
        return cls._create_dir_if_needed(cls.get_plugins_dir())

    @classmethod
    def get_db_dir(cls, test_db=False):
        """Return the configured directory of where the database will be stored."""
        if test_db:
            base = cls.temp_dir
        else:
            base = cls.get_base_dir()
        return cls._create_dir_if_needed(base + os.sep + cls.directories['db_dir'])

    @classmethod
    def get_db_params(cls, test_db=False):
        """Return the database configuration."""
        db_type = cls.get_db_type()
        db_params = {
            'db_type' : db_type
        }
        if db_type == 'sqlite':
            db_params['db_path'] = cls.get_db_dir(test_db)
        elif db_type == "mysql":
            db_params['db_type'] = 'mysql'
            db_params['db_username'] = cls.get_db_user()
            db_params['db_password'] = cls.get_db_password()
            db_params['db_host'] = cls.get_db_host()
        return DbParams(**db_params)

    @classmethod
    def get_metric(cls):
        """Return the unit system (metric, statute) that is configured."""
        return cls.config['metric']

    @classmethod
    def device_settings_dir(cls, mount_dir):
        """Return the full path to the settings file on a mounted device."""
        return mount_dir + os.sep + cls.device_directories['base'] + os.sep + cls.device_directories['settings']

    @classmethod
    def device_monitoring_dir(cls, mount_dir):
        """Return the full path to the monitoring files on a mounted device."""
        return mount_dir + os.sep + cls.device_directories['base'] + os.sep + cls.device_directories['monitoring']

    @classmethod
    def device_sleep_dir(cls, mount_dir):
        """Return the full path to the sleep files on a mounted device."""
        return mount_dir + os.sep + cls.device_directories['base'] + os.sep + cls.device_directories['sleep']

    @classmethod
    def device_activities_dir(cls, mount_dir):
        """Return the full path to the activities files on a mounted device."""
        return mount_dir + os.sep + cls.device_directories['base'] + os.sep + cls.device_directories['activities']

    @classmethod
    def get_graphs(cls, key):
        """Return a graph config item."""
        return cls.graphs.get(key)

    @classmethod
    def graphs_activity_config(cls, activity, key):
        """Return a config value for the graphing capability given it's key name."""
        activity = cls.graphs.get(activity)
        if activity is not None:
            return activity.get(key)



#!/usr/bin/env python3

"""Class running a checkup against the DB data."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import sys
import logging
import argparse
from datetime import datetime, time, timedelta

import fitfile

from garmindb.garmindb import GarminDb, Attributes, Device, DeviceInfo, DailySummary, ActivitiesDb, Activities, StepsActivities
from garmindb import ConfigManager, format_version


logging.basicConfig(filename='graphs.log', filemode='w', level=logging.INFO)
logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class CheckUp(object):
    """Class running a checkup against the DB data."""

    def __init__(self, debug):
        """Return an instance of the CheckUp class."""
        self.db_params = ConfigManager.get_db_params()
        self.debug = debug
        self.garmin_db = GarminDb(self.db_params)
        self.measurement_system = Attributes.measurements_type(self.garmin_db)
        self.unit_strings = fitfile.units.unit_strings[self.measurement_system]

    def goals(self):
        """Do a checkup of th euser's goals."""
        look_back_days = ConfigManager.checkup.get('look_back_days')
        end_ts = datetime.now()
        start_ts = end_ts - timedelta(days=look_back_days)
        results = DailySummary.get_for_period(self.garmin_db, start_ts, end_ts)
        step_goal_days = 0
        step_goal_days_in_week = 0
        floors_goal_days = 0
        floor_goal_days_in_week = 0
        days_in_week = 0
        intensity_time = time.min
        intensity_time_goal = time.min
        intensity_weeks = 0
        intensity_goal_weeks = 0
        for result in results:
            if result.day.weekday() == 0:
                days_in_week = 0
                step_goal_days_in_week = 0
                floor_goal_days_in_week = 0
                intensity_time = time.min
                intensity_time_goal = time.min
            days_in_week += 1
            if result.steps_goal_percent >= 100:
                step_goal_days += 1
                step_goal_days_in_week += 1
            else:
                logger.debug('Steps: goal not met on %s', result.day)
            if result.floors_goal_percent >= 100:
                floors_goal_days += 1
                floor_goal_days_in_week += 1
            else:
                logger.debug('Floors: goal not met on %s', result.day)
            intensity_time = fitfile.conversions.add_time(intensity_time, result.intensity_time)
            intensity_time_goal = fitfile.conversions.add_time(intensity_time_goal, result.intensity_time_goal)
            if result.day.weekday() == 6:
                if days_in_week == 7:
                    intensity_weeks += 1
                    if step_goal_days_in_week < days_in_week:
                        logger.info('Steps: goal not met %d days for week ending in %s', days_in_week - step_goal_days_in_week, result.day)
                    if floor_goal_days_in_week < days_in_week:
                        logger.info('Floors: goal not met %d days for week ending in %s', days_in_week - floor_goal_days_in_week, result.day)
                    if intensity_time >= intensity_time_goal:
                        intensity_goal_weeks += 1
                    else:
                        logger.info('Intensity mins: goal not met for week ending in %s', result.day)
        logger.info('Summary:')
        logger.info('Steps: met goal %d of last %d days', step_goal_days, look_back_days)
        logger.info('Floors: met goal %d of last %d days', floors_goal_days, look_back_days)
        logger.info('Intensity mins: met goal %d of last %d weeks', intensity_goal_weeks, intensity_weeks)

    def __activity_string(self, activity_db, activity):
        if activity.is_steps_activity():
            steps_activity = StepsActivities.get(activity_db, activity.activity_id)
            return ('%s: "%s" %.2f %s in %s pace: %s %s speed: %.2f %s' %
                    (activity.start_time, activity.name, activity.distance, self.unit_strings[fitfile.units.UnitTypes.distance_long], activity.elapsed_time,
                     steps_activity.avg_pace, self.unit_strings[fitfile.units.UnitTypes.pace], activity.avg_speed,
                     self.unit_strings[fitfile.units.UnitTypes.speed]))
        return '%s: "%s" %s in %s (%s)' % (activity.start_time, activity.name, activity.distance, activity.elapsed_time, activity.avg_speed)

    def activity_course(self, course_id):
        """Run a checkup on all activities matching the course_id."""
        activity_db = ActivitiesDb(self.db_params, self.debug)
        activities = Activities.get_by_course_id(activity_db, course_id)
        activities_count = len(activities)
        fastest_activity = Activities.get_fastest_by_course_id(activity_db, course_id)
        slowest_activity = Activities.get_slowest_by_course_id(activity_db, course_id)
        logger.info('Matching Activities: %d', activities_count)
        logger.info('  first: %s', self.__activity_string(activity_db, activities[0]))
        logger.info('  lastest: %s', self.__activity_string(activity_db, activities[-1]))
        logger.info('  fastest: %s', self.__activity_string(activity_db, fastest_activity))
        logger.info('  slowest: %s', self.__activity_string(activity_db, slowest_activity))

    def battery_status(self):
        """Check for devices with low battery status."""
        devices = Device.get_all(self.garmin_db)
        for device in devices:
            battery_level = DeviceInfo.get_col_latest_where(self.garmin_db, DeviceInfo.battery_status,
                                                            [DeviceInfo.serial_number == device.serial_number, DeviceInfo.battery_status != fitfile.field_enums.BatteryStatus.invalid])
            if battery_level is fitfile.field_enums.BatteryStatus.low:
                logger.info("Device %s %s (%s) has a low battery", device.manufacturer, device.product, device.serial_number)


def main(argv):
    """Run a data checkup of the user's choice."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", help="print the program's version", action='version', version=format_version(sys.argv[0]))
    parser.add_argument("-t", "--trace", help="Turn on debug tracing", type=int, default=0)
    checks_group = parser.add_argument_group('Checks')
    checks_group.add_argument("-b", "--battery", help="Check for low battery levels.", action="store_true", default=False)
    checks_group.add_argument("-c", "--course", help="Show statistics from all workouts for a single course.", type=int, default=None)
    checks_group.add_argument("-g", "--goals", help="Run a checkup on the user\'s goals.", action="store_true", default=False)
    checks_group.add_argument("-a", "--all", help="Run a checkup on all of the the user\'s stats.", action="store_true", default=False)
    args = parser.parse_args()

    checkup = CheckUp(args.trace)
    if args.all or args.battery:
        checkup.battery_status()
    if args.course:
        checkup.activity_course(args.course)
    if args.all or args.goals:
        checkup.goals()


if __name__ == "__main__":
    main(sys.argv[1:])



"""Test FIT file parsing."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging

from idbutils import FileProcessor

from garmindb import Tcx


root_logger = logging.getLogger()
handler = logging.FileHandler('tcx_file.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestTcxFile(unittest.TestCase):
    """Class for testing TCX file parsing."""

    @classmethod
    def setUpClass(cls):
        cls.file_path = 'test_files/tcx'
        cls.tcx_filename_regex = r'.*\.tcx'

    def check_activity_file(self, filename):
        tcx = Tcx()
        tcx.read(filename)
        (manufacturer, product) = tcx.get_manufacturer_and_product()
        logger.info("%s: sport %r end_time %r start_time %r manufacturer %r product %r serial_number %r laps %r distance %r calories %r start %r end %r "
                    "hr %r, %r max speed %r cadence %r, %r",
                    filename, tcx.sport, tcx.end_time, tcx.start_time, manufacturer, product, tcx.serial_number, tcx.lap_count,
                    tcx.distance, tcx.calories, tcx.start_loc, tcx.end_loc, tcx.hr_avg, tcx.hr_max, tcx.speed_max, tcx.cadence_avg, tcx.cadence_max)
        self.assertGreater(tcx.end_time, tcx.start_time)
        self.assertGreater(tcx.lap_count, 0)

    def test_parse_tcx(self):
        file_names = FileProcessor.dir_to_files(self.file_path, self.tcx_filename_regex, False)
        for file_name in file_names:
            self.check_activity_file(file_name)


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""Version information for the application."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


python_required = (3, 0, 0)
python_tested = (3, 9, 5)
version_info = (2, 0, 4)
prerelease = True


def version_string():
    """Return a version string for a version tuple."""
    return '.'.join(str(digit) for digit in version_info)



"""Test profile file parsing."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging

import fitfile

from garmindb import ConfigManager, GarminProfile
from garmindb.garmindb import GarminDb, Attributes


root_logger = logging.getLogger()
handler = logging.FileHandler('profile_file.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestProfileFile(unittest.TestCase):
    """Class for testing profile JSON file parsing."""

    @classmethod
    def setUpClass(cls):
        cls.file_path = 'test_files'

    def test_parse_uprofile(self):
        db_params = ConfigManager.get_db_params(test_db=True)
        gp = GarminProfile(db_params, self.file_path, debug=2)
        if gp.file_count() > 0:
            gp.process()
        gdb = GarminDb(db_params)
        measurement_system = Attributes.measurements_type(gdb)
        self.assertEqual(measurement_system, fitfile.field_enums.DisplayMeasure.statute,
                         'DisplayMeasure expected %r found %r' % (fitfile.field_enums.DisplayMeasure.statute, measurement_system))


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""A database and database objects for storing health data from a Microsoft Health."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

# flake8: noqa

from .mshealth_db import MSHealthDb, Attributes, DaysSummary, MSVaultWeight
from .import_csv import MSHealthData, MSVaultData
from .analyze import Analyze



"""Test activities db."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import unittest
import logging

import fitfile

from garmindb import GarminActivitiesFitData, GarminTcxData, GarminJsonSummaryData, GarminJsonDetailsData, ActivityFitFileProcessor, ConfigManager, PluginManager
from garmindb.garmindb import GarminDb, Device, File, DeviceInfo
from garmindb.garmindb import ActivitiesDb, Activities, ActivityLaps, ActivityRecords, StepsActivities, PaddleActivities, CycleActivities

from test_db_base import TestDBBase


root_logger = logging.getLogger()
root_logger.addHandler(logging.FileHandler('activities_db.log', 'w'))
root_logger.setLevel(logging.DEBUG)

logger = logging.getLogger(__name__)

do_fit_import_test = True
do_tcx_import_tests = True
do_summary_import_tests = True
do_details_import_tests = True
do_multiple_import_tests = True


class TestActivitiesDb(TestDBBase, unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.garmin_act_db = ActivitiesDb(ConfigManager.get_db_params())
        table_dict = {
            'activities_table' : Activities,
            'activity_laps_table' : ActivityLaps,
            'activity_records_table' : ActivityRecords,
            'run_activities_table' : StepsActivities,
            'paddle_activities_table' : PaddleActivities,
            'cycle_activities_table' : CycleActivities,
        }
        super().setUpClass(cls.garmin_act_db, table_dict, {Activities : [Activities.activity_id]})
        cls.test_db_params = ConfigManager.get_db_params(test_db=True)
        cls.plugin_manager = PluginManager(ConfigManager.get_or_create_plugins_dir(), cls.test_db_params)
        cls.test_mon_db = GarminDb(cls.test_db_params)
        cls.test_act_db = ActivitiesDb(cls.test_db_params, debug_level=1)
        cls.measurement_system = fitfile.field_enums.DisplayMeasure.statute
        print(f"db params {repr(cls.test_db_params)}")

    def test_garmin_act_db_tables_exists(self):
        self.assertGreater(Activities.row_count(self.garmin_act_db), 0)
        self.assertGreater(ActivityLaps.row_count(self.garmin_act_db), 0)
        self.assertGreater(ActivityRecords.row_count(self.garmin_act_db), 0)
        self.assertGreater(StepsActivities.row_count(self.garmin_act_db), 0)
        self.assertGreater(PaddleActivities.row_count(self.garmin_act_db), 0)
        self.assertGreater(CycleActivities.row_count(self.garmin_act_db), 0)

    def check_activities_fields(self, fields_list):
        self.check_not_none_cols(self.test_act_db, {Activities : fields_list})

    def check_activities_field_value(self, field, min_value, max_value):
        field_max = Activities.get_col_max(self.test_act_db, field)
        if field_max is not None:
            self.assertLessEqual(field_max, max_value)
        field_min = Activities.get_col_min(self.test_act_db, field)
        if field_min is not None:
            self.assertGreaterEqual(field_min, min_value)

    def check_sport(self, activity):
        sport = fitfile.Sport.from_string(activity.sport)
        self.assertIsInstance(sport, fitfile.Sport, f'sport ({type(sport)}) from {repr(activity)}')
        self.assertEqual(activity.sport, sport.name)
        sub_sport = fitfile.SubSport.from_string(activity.sub_sport)
        self.assertIsInstance(sub_sport, fitfile.SubSport, f'sub_sport ({type(sub_sport)}) from {repr(activity)}')
        self.assertEqual(activity.sub_sport, sub_sport.name)

    def check_activities(self):
        for activity in Activities.get_all(self.test_act_db):
            self.check_sport(activity)
        self.check_col_type(self.test_act_db, Activities, Activities.activity_id, int)

    def __fit_file_import(self):
        gfd = GarminActivitiesFitData('test_files/fit/activity', latest=False, measurement_system=self.measurement_system, debug=2)
        self.gfd_file_count = gfd.file_count()
        if gfd.file_count() > 0:
            gfd.process_files(ActivityFitFileProcessor(self.test_db_params, self.plugin_manager))

    def fit_file_import(self):
        self.profile_function('fit_activities_import', self.__fit_file_import)
        self.check_db_tables_exists(self.test_mon_db, {'device_table' : Device})
        self.check_db_tables_exists(self.test_mon_db, {'file_table' : File, 'device_info_table' : DeviceInfo}, self.gfd_file_count)

    def summary_json_file_import(self):
        gjsd = GarminJsonSummaryData(self.test_db_params, 'test_files/json/activity/summary', latest=False, measurement_system=self.measurement_system, debug=2)
        if gjsd.file_count() > 0:
            gjsd.process()

    def details_json_file_import(self, delete_db=True):
        gjsd = GarminJsonDetailsData(self.test_db_params, 'test_files/json/activity/details', latest=False, measurement_system=self.measurement_system, debug=2)
        if gjsd.file_count() > 0:
            gjsd.process()

    def tcx_file_import(self):
        ActivitiesDb.delete_db(self.test_db_params)
        gtd = GarminTcxData('test_files/tcx', latest=False, measurement_system=self.measurement_system, debug=2)
        if gtd.file_count() > 0:
            gtd.process_files(self.test_db_params)

    #
    # The actual tests
    #
    @unittest.skipIf(not do_fit_import_test, "Skipping fit import test")
    def test_fit_file_import(self):
        ActivitiesDb.delete_db(self.test_db_params)
        self.fit_file_import()
        self.check_activities_fields([Activities.start_time, Activities.stop_time, Activities.elapsed_time])
        self.check_activities()
        self.check_activities_field_value(Activities.avg_speed, 0, 50)

    @unittest.skipIf(not do_tcx_import_tests, "Skipping tcx import test")
    def test_tcx_file_import(self):
        ActivitiesDb.delete_db(self.test_db_params)
        self.tcx_file_import()
        self.check_activities_fields([Activities.sport, Activities.laps])

    @unittest.skipIf(not do_summary_import_tests, "Skipping summary import test")
    def test_summary_json_file_import(self):
        ActivitiesDb.delete_db(self.test_db_params)
        self.summary_json_file_import()
        self.check_activities_fields([Activities.name, Activities.type, Activities.sport, Activities.sub_sport])
        self.check_activities_field_value(Activities.avg_speed, 0, 50)

    @unittest.skipIf(not do_details_import_tests, "Skipping details import test")
    def test_details_json_file_import(self):
        ActivitiesDb.delete_db(self.test_db_params)
        self.details_json_file_import()

    @unittest.skipIf(not do_multiple_import_tests, "Skipping multiple import test")
    def test_file_import(self):
        root_logger.info("test_file_import: %r", self.test_db_params)
        self.summary_json_file_import()
        self.details_json_file_import()
        self.tcx_file_import()
        self.fit_file_import()
        self.check_activities_fields([Activities.start_time, Activities.stop_time, Activities.elapsed_time])


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""A script for importing CSV formatted Microsoft Health export data."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import sys
import re
import logging
from tqdm import tqdm

from idbutils import CsvImporter
from idbutils import FileProcessor

from .mshealth_db import MSHealthDb, MSVaultWeight, DaysSummary

logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))


class MSHealthData(object):
    """A classs for importing CSV formatted Microsoft Health export data."""

    cols_map = {
        'Date': ('day', CsvImporter.map_ymd_date),
        'Floors_Climbed': ('floors', CsvImporter.map_identity),
        'Steps': ('steps', CsvImporter.map_integer),
        'HR_Highest': ('hr_max', CsvImporter.map_integer),
        'HR_Lowest': ('hr_min', CsvImporter.map_integer),
        'HR_Average': ('hr_avg', CsvImporter.map_integer),
        'Calories': ('calories', CsvImporter.map_integer),
        'Active_Hours': ('active_hours', CsvImporter.map_integer),
        'Total_Seconds_All_Activities': ('activity_secs', CsvImporter.map_integer),
        'Total_Calories_All_Activities': ('activity_calories', CsvImporter.map_integer),
        'Exercise_Events': ('exercise_events', CsvImporter.map_integer),
        'Exercise_Total_Calories': ('exercise_calories', CsvImporter.map_integer),
        'Exercise_Total_Seconds': ('exercise_secs', CsvImporter.map_integer),
        'Total_Miles_Moved': ('miles_moved', CsvImporter.map_float),
        'Sleep_Events': ('sleep_events', CsvImporter.map_integer),
        'Sleep_Total_Calories': ('sleep_calories', CsvImporter.map_integer),
        'Total_Seconds_Slept': ('sleep_secs', CsvImporter.map_integer),
        'Walk_Events': ('walk_events', CsvImporter.map_integer),
        'Walk_Total_Seconds': ('walk_secs', CsvImporter.map_integer),
        'Walk_Total_Calories': ('workout_calories', CsvImporter.map_integer),
        'Total_Miles_Walked': ('miles_walked', CsvImporter.map_float),
        'Run_Events': ('run_ewvents', CsvImporter.map_integer),
        'Run_Total_Calories': ('run_calories', CsvImporter.map_integer),
        'Run_Total_Seconds': ('run_secs', CsvImporter.map_integer),
        'Total_Miles_Run': ('miles_run', CsvImporter.map_float),
        'Total_Miles_Golfed': ('miles_golfed', CsvImporter.map_float),
        'Golf_Total_Calories': ('golf_calories', CsvImporter.map_integer),
        'Golf_Events': ('golf_events', CsvImporter.map_integer),
        'Golf_Total_Seconds': ('golf_secs', CsvImporter.map_integer),
        'Total_Miles_Biked': ('miles_biked', CsvImporter.map_float),
        'UV_Exposure_Minutes': ('uv_mins', CsvImporter.map_integer),
        'Bike_Total_Seconds': ('bike_secs', CsvImporter.map_integer),
        'Bike_Total_Calories': ('bike_calories', CsvImporter.map_integer),
        'Bike_Events': ('bike_events', CsvImporter.map_integer),
        'Guided_Workout_Events': ('guided_workout_events', CsvImporter.map_integer),
        'Guided_Workout_Total_Calories': ('guided_workout_calories', CsvImporter.map_integer),
        'Guided_Workout_Total_Seconds': ('guided_workout_secs', CsvImporter.map_integer),
    }

    def __init__(self, input_file, input_dir, db_params, metric, debug):
        """Return an instance of MSHealthData given an input file or files and information on the databse to put it in."""
        self.metric = metric
        self.mshealth_db = MSHealthDb(db_params, debug)
        if input_file:
            self.file_names = FileProcessor.match_file(input_file, r'Daily_Summary_.*\.csv')
        if input_dir:
            self.file_names = FileProcessor.dir_to_files(input_dir, r'Daily_Summary_.*\.csv')

    def file_count(self):
        """Return the number of files that will be processed."""
        return len(self.file_names)

    def __write_entry(self, db_entry):
        DaysSummary.insert_or_update(self.mshealth_db, db_entry)

    def process_files(self):
        """Import files into the databse."""
        for file_name in tqdm(self.file_names, unit='files'):
            logger.info("Processing file: " + file_name)
            csvimporter = CsvImporter(file_name, self.cols_map, self.__write_entry)
            csvimporter.process_file(not self.metric)


class MSVaultData(object):
    """A class for importing CSV formatted Microsoft Health Vault export data."""

    def __init__(self, input_file, input_dir, db_params, metric, debug):
        """Return an instance of MSVaultData given an input file or files and information on the databse to put it in."""
        self.metric = metric
        self.mshealth_db = MSHealthDb(db_params, debug)
        self.cols_map = {
            'Date': ('timestamp', CsvImporter.map_mdy_date),
            'Weight': ('weight', MSVaultData.__map_weight),
        }
        if input_file:
            self.file_names = FileProcessor.match_file(input_file, r'HealthVault_Weight_.*\.csv')
        if input_dir:
            self.file_names = FileProcessor.dir_to_files(input_dir, r'HealthVault_Weight_.*\.csv')

    def file_count(self):
        """Return the number of files that will be processed."""
        return len(self.file_names)

    def __write_entry(self, db_entry):
        try:
            MSVaultWeight.insert_or_update(self.mshealth_db, MSVaultWeight.intersection(db_entry))
        except Exception as e:
            logger.error('Failed to save %r to db: %s', db_entry, e)

    def process_files(self):
        """Import files into the databse."""
        for file_name in tqdm(self.file_names, unit='files'):
            logger.info("Processing file: " + file_name)
            csvimporter = CsvImporter(file_name, self.cols_map, self.__write_entry)
            csvimporter.process_file(not self.metric)

    @classmethod
    def __map_weight(cls, metric, value):
        m = re.search(r'(\d{2,3}\.\d{2}) .*', value)
        if m:
            logger.debug("Matched weight: " + m.group(1))
            return float(m.group(1))
        else:
            logger.debug("Unmatched weight: " + value)
            return None



"""Base classes for plugins."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging

import idbutils

from .garmindb import ActivitiesDb, Activities


logger = logging.getLogger(__file__)


class PluginManager(idbutils.PluginManager):
    """Loads python file based plugins that extend GarminDb."""

    def __init__(self, plugin_dir, db_params):
        """Load python file based plugins from plugin_dir."""
        logger.info("Loading GarminDb plugins from %s", plugin_dir)
        super().__init__(plugin_dir, {'db_params': db_params})

    def get_activity_file_processors(self, fit_file):
        """Return a dict of all plugins that handle FIT file messages."""
        result = {}
        for plugin_name, plugin in self.plugins.items():
            if plugin.matches_activity_file(fit_file):
                logger.info("Plugin %s matches file %s", plugin_name, fit_file)
                plugin.init_activity(ActivitiesDb, Activities)
                result[plugin_name] = plugin
        return result



"""Test Tcx file creation."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging
import datetime

from fitfile import GarminProduct, Distance, Speed, Sport
from idbutils import Location

from garmindb import Tcx


root_logger = logging.getLogger()
handler = logging.FileHandler('loop.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestLoop(unittest.TestCase):
    """Class for testing Tcx file creation."""

    @classmethod
    def setUpClass(cls):
        cls.file_path = 'test_files/tcx'
        cls.tcx_filename_regex = r'.*\.tcx'

    def test_loop(self):
        sport = Sport.boating.name
        start_time = datetime.datetime.now()
        stop_time = start_time + datetime.timedelta(hours=1)
        product = GarminProduct.Fenix.name
        serial_number = 123412341234
        version = (1, 2, 3, 4)
        lap_distance = Distance.from_meters(10)
        lap_calories = 100
        position1 = Location(1, 2)
        position2 = Location(3, 4)
        record_alititude = Distance.from_meters(100)
        record_hr = 100
        record_speed = Speed.from_mph(6.0)
        # create a TCX XML tree
        tcx = Tcx()
        tcx.create(sport, start_time)
        tcx.add_creator(product, serial_number, version=version)
        track = tcx.add_lap(start_time, stop_time, lap_distance, lap_calories)
        tcx.add_point(track, start_time, position1, record_alititude, record_hr, record_speed)
        tcx.add_point(track, stop_time, position2, record_alititude, record_hr, record_speed)
        # now read it back
        tcx.update()
        self.assertEqual(tcx.sport, sport)
        self.assertEqual(tcx.start_time, start_time)
        self.assertEqual(tcx.end_time, stop_time)
        self.assertEqual(tcx.calories, lap_calories)
        self.assertEqual(tcx.distance, lap_distance)
        self.assertEqual(tcx.duration, 60 * 60)
        self.assertEqual(tcx.start_loc, position1)
        self.assertEqual(tcx.end_loc, position2)
        self.assertEqual(tcx.creator_product, product)
        self.assertEqual(int(tcx.creator_serialnumber), serial_number)
        self.assertEqual(tcx.creator_version, version)
        logger.info('hr avg: %f', tcx.hr_avg)
        logger.info('hr max: %f', tcx.hr_max)


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""Objects representing a database and database objects summarizing health data."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging
from sqlalchemy import Column, Date

import idbutils

from .summary_base import SummaryBase


logger = logging.getLogger(__name__)

SummaryDb = idbutils.DB.create('summary', 7, "Database for storing summarizing health data.")
Summary = idbutils.DbObject.create('summary', SummaryDb, 1, base=idbutils.KeyValueObject)


class YearsSummary(SummaryDb.Base, SummaryBase):
    """Object representing summarized monthly health data."""

    __tablename__ = 'years_summary'

    db = SummaryDb
    table_version = 3
    view_version = SummaryBase.view_version

    first_day = Column(Date, primary_key=True)

    @classmethod
    def create_view(cls, db):
        """Create the default database view for the table."""
        cls.create_years_view(db)


class MonthsSummary(SummaryDb.Base, SummaryBase):
    """Object representing summarized monthly health data."""

    __tablename__ = 'months_summary'

    db = SummaryDb
    table_version = 3
    view_version = SummaryBase.view_version

    first_day = Column(Date, primary_key=True)

    @classmethod
    def create_view(cls, db):
        """Create the default database view for the table."""
        cls.create_months_view(db)


class WeeksSummary(SummaryDb.Base, SummaryBase):
    """Object representing summarized weekly health data."""

    __tablename__ = 'weeks_summary'

    db = SummaryDb
    table_version = 3
    view_version = SummaryBase.view_version

    first_day = Column(Date, primary_key=True)

    @classmethod
    def create_view(cls, db):
        """Create the default database view for the table."""
        cls.create_weeks_view(db)


class DaysSummary(SummaryDb.Base, SummaryBase):
    """Object representing summarized daily health data."""

    __tablename__ = 'days_summary'

    db = SummaryDb
    table_version = 3
    view_version = SummaryBase.view_version

    day = Column(Date, primary_key=True)

    @classmethod
    def create_view(cls, db):
        """Create the default database view for the table."""
        cls.create_days_view(db)



"""Objects representing a database and databse objects for storing activities data."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging
import datetime
from sqlalchemy import Column, String, Float, Integer, DateTime, Time, ForeignKey, PrimaryKeyConstraint, desc, literal_column
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship
from sqlalchemy.ext.hybrid import hybrid_property

import idbutils


logger = logging.getLogger(__name__)

ActivitiesDb = idbutils.DB.create('garmin_activities', 13, "Database for storing activities data.")


class ActivitiesLocationSegment(idbutils.DbObject):
    """Object representing a databse object for storing location segnment from an activity."""

    # degrees
    start_lat = Column(Float)
    start_long = Column(Float)
    stop_lat = Column(Float)
    stop_long = Column(Float)

    @hybrid_property
    def start_loc(self):
        """Return the starting location of activity segment as a Location instance."""
        return idbutils.Location(self.start_lat, self.start_long)

    @start_loc.setter
    def start_loc(self, start_location):
        self.start_lat = start_location.lat_deg
        self.start_long = start_location.long_deg

    @hybrid_property
    def stop_loc(self):
        """Return the ending location of activity segment as a Location instance."""
        return idbutils.Location(self.stop_lat, self.stop_long)

    @stop_loc.setter
    def stop_loc(self, stop_location):
        self.stop_lat = stop_location.lat_deg
        self.stop_long = stop_location.long_deg


class Activities(ActivitiesDb.Base, ActivitiesLocationSegment):
    """Class represents a databse table that contains data about recorded activities."""

    __tablename__ = 'activities'

    db = ActivitiesDb
    table_version = 3

    activity_id = Column(String, primary_key=True)
    name = Column(String)
    description = Column(String)
    type = Column(String)
    #
    course_id = Column(Integer)
    #
    start_time = Column(DateTime)
    stop_time = Column(DateTime)
    elapsed_time = Column(Time, nullable=False, default=datetime.time.min)
    moving_time = Column(Time, nullable=False, default=datetime.time.min)
    #
    sport = Column(String)
    sub_sport = Column(String)
    # kms or miles
    distance = Column(Float)
    #
    cycles = Column(Float)
    #
    laps = Column(Integer)
    # beats per minute
    avg_hr = Column(Integer)
    max_hr = Column(Integer)
    # breaths per minute
    avg_rr = Column(Float)
    max_rr = Column(Float)
    calories = Column(Integer)
    avg_cadence = Column(Integer)
    max_cadence = Column(Integer)
    # kmph or mph
    avg_speed = Column(Float)
    max_speed = Column(Float)
    # feet or meters
    ascent = Column(Float)
    descent = Column(Float)
    # C or F
    max_temperature = Column(Float)
    min_temperature = Column(Float)
    avg_temperature = Column(Float)

    training_effect = Column(Float)
    anaerobic_training_effect = Column(Float)

    def is_steps_activity(self):
        """Return if the activity is a steps based activity."""
        return self.sport in ['walking', 'running', 'hiking']

    @classmethod
    def get_by_course_id(cls, db, course_id):
        """Return all activities records for activities with the matching course_id."""
        with db.managed_session() as session:
            return session.query(cls).filter(cls.course_id == course_id).order_by(cls.start_time).all()

    @classmethod
    def get_fastest_by_course_id(cls, db, course_id):
        """Return an activities record for the activity with the matching course_id with the fastest speed."""
        with db.managed_session() as session:
            return session.query(cls).filter(cls.course_id == course_id).order_by(desc(cls.avg_speed)).limit(1).one_or_none()

    @classmethod
    def get_slowest_by_course_id(cls, db, course_id):
        """Return an activities record for the activity with the matching course_id with the slowest speed."""
        with db.managed_session() as session:
            return session.query(cls).filter(cls.course_id == course_id).order_by(cls.avg_speed).limit(1).one_or_none()

    @classmethod
    def get_stats(cls, session, start_ts, end_ts):
        """Return a dict of stats for the time range."""
        stats = {
            'activities'            : cls.s_row_count_for_period(session, start_ts, end_ts),
            'activities_calories'   : cls.s_get_col_sum(session, cls.calories, start_ts, end_ts),
            'activities_distance'   : cls.s_get_col_sum(session, cls.distance, start_ts, end_ts),
        }
        return stats


class ActivityLaps(ActivitiesDb.Base, ActivitiesLocationSegment):
    """Class that holds data for an activity lap."""

    __tablename__ = 'activity_laps'

    db = ActivitiesDb
    table_version = 3

    activity_id = Column(String, ForeignKey('activities.activity_id'))
    lap = Column(Integer)
    #
    start_time = Column(DateTime)
    stop_time = Column(DateTime)
    elapsed_time = Column(Time, nullable=False, default=datetime.time.min)
    moving_time = Column(Time, nullable=False, default=datetime.time.min)
    # kms or miles
    distance = Column(Float)
    cycles = Column(Float)
    # beats per minute
    avg_hr = Column(Integer)
    max_hr = Column(Integer)
    # breaths per minute
    avg_rr = Column(Float)
    max_rr = Column(Float)
    calories = Column(Integer)
    avg_cadence = Column(Integer)
    max_cadence = Column(Integer)
    # kmph or mph
    avg_speed = Column(Float)
    max_speed = Column(Float)
    # feet or meters
    ascent = Column(Float)
    descent = Column(Float)
    # C or F
    max_temperature = Column(Float)
    min_temperature = Column(Float)
    avg_temperature = Column(Float)

    __table_args__ = (PrimaryKeyConstraint("activity_id", "lap"),)

    @classmethod
    def s_get_activity(cls, session, activity_id):
        """Return all laps for a given activity_id."""
        return session.query(cls).filter(cls.activity_id == activity_id).all()

    @hybrid_property
    def start_loc(self):
        """Return the lap start location."""
        return idbutils.Location(self.start_lat, self.start_long)

    @start_loc.setter
    def start_loc(self, start_location):
        self.start_lat = start_location.lat_deg
        self.start_long = start_location.long_deg


class ActivityRecords(ActivitiesDb.Base, idbutils.DbObject):
    """Encapsilates record for a single point in time from an activity."""

    __tablename__ = 'activity_records'

    db = ActivitiesDb
    table_version = 3

    activity_id = Column(String, ForeignKey('activities.activity_id'))
    record = Column(Integer)
    timestamp = Column(DateTime)
    position_lat = Column(Float)    # degrees
    position_long = Column(Float)   # degrees
    distance = Column(Float)
    cadence = Column(Integer)
    altitude = Column(Float)
    hr = Column(Integer)            # beats per minute
    rr = Column(Float)              # breaths per minute
    altitude = Column(Float)        # feet or meters
    speed = Column(Float)           # kmph or mph
    temperature = Column(Float)     # C or F

    __table_args__ = (PrimaryKeyConstraint("activity_id", "record"),)

    @classmethod
    def s_get_activity(cls, session, activity_id):
        """Return all records for a given activity_id."""
        return session.query(cls).filter(cls.activity_id == activity_id).all()

    @hybrid_property
    def position(self):
        """Return the location where the record was recorded."""
        return idbutils.Location(self.position_lat, self.position_long)

    @position.setter
    def position(self, location):
        self.position_lat = location.lat_deg
        self.position_long = location.long_deg


class SportActivities(idbutils.DbObject):
    """Base class for all sport based activity tables."""

    @declared_attr
    def activity_id(cls):
        return Column(String, ForeignKey(Activities.activity_id), primary_key=True)

    @declared_attr
    def activity(cls):
        return relationship("Activities")

    @classmethod
    def _create_activity_view(cls, db, selectable):
        """Create a database view for a activity type."""
        view_name = cls._get_default_view_name()
        logger.debug("Creating activity view %s if needed.", view_name)
        cls.create_join_view(db, view_name, selectable, Activities, order_by=Activities.start_time.desc())

    @classmethod
    def _create_sport_view(cls, db, selectable, sport):
        """Create a database view for a sport based activity type."""
        filter = literal_column(f'{Activities.sport} == "{sport}"')
        cls.create_join_view(db, f'{sport}_activities_view', selectable, Activities, filter, Activities.start_time.desc())

    @classmethod
    def _create_course_view(cls, db, selectable, course_id):
        filter = literal_column(f'{Activities.course_id} == {course_id}')
        cls.create_join_view(db, f'course_{course_id}_view', selectable, Activities, filter, Activities.start_time.desc())

    @classmethod
    def create_view(cls, db):
        """Create a database view for a sport based activity type."""
        cls._create_activity_view(db, cls._view_selectable())

    @classmethod
    def google_map_loc(cls, label):
        """Return a literal column composed of a google map URL for either the start or stop location off the activity."""
        return literal_column(idbutils.Location.google_maps_url('activities.%s_lat' % label, 'activities.%s_long' % label) + ' AS %s_loc' % label)


class StepsActivities(ActivitiesDb.Base, SportActivities):
    """Step based activity table."""

    __tablename__ = 'steps_activities'

    db = ActivitiesDb
    table_version = 3
    view_version = 5

    steps = Column(Integer)
    # pace in mins/mile
    avg_pace = Column(Time, nullable=False, default=datetime.time.min)
    avg_moving_pace = Column(Time, nullable=False, default=datetime.time.min)
    max_pace = Column(Time, nullable=False, default=datetime.time.min)
    # steps per minute
    avg_steps_per_min = Column(Integer)
    max_steps_per_min = Column(Integer)
    # m or ft
    avg_step_length = Column(Float)
    # %
    avg_vertical_ratio = Column(Float)
    # m or ft
    avg_vertical_oscillation = Column(Float)
    # left % of left right balance
    avg_gct_balance = Column(Float)
    # ground contact time in ms
    avg_ground_contact_time = Column(Time, nullable=False, default=datetime.time.min)
    avg_stance_time_percent = Column(Float)
    vo2_max = Column(Float)

    @classmethod
    def _view_selectable(cls, include_sport=False, include_subsport=False, include_type=False, include_course=False, include_rr=False, include_running_dynamics=False):
        # The query fails to genarate sql when using the func.round clause.
        selectable = [
            Activities.activity_id.label('activity_id'),
            Activities.name.label('name'),
            Activities.description.label('description')
        ]
        if include_sport:
            selectable.append(Activities.sport.label('sport'))
        if include_subsport:
            selectable.append(Activities.sub_sport.label('sub_sport'))
        if include_type:
            selectable.append(Activities.type.label('type'))
        if include_course:
            selectable.append(Activities.course_id.label('course_id'))
        selectable += [
            Activities.start_time.label('start_time'),
            Activities.stop_time.label('stop_time'),
            Activities.elapsed_time.label('elapsed_time'),
            cls.round_ext_col(Activities, 'distance'),
            cls.steps.label('steps'),
            cls.avg_pace .label('avg_pace'),
            cls.avg_moving_pace.label('avg_moving_pace'),
            cls.max_pace.label('max_pace'),
            cls.round_col('avg_steps_per_min'),
            cls.round_col('max_steps_per_min'),
            Activities.avg_hr.label('avg_hr'),
            Activities.max_hr.label('max_hr')
        ]
        if include_rr:
            selectable += [Activities.avg_rr.label('avg_rr'), Activities.max_rr.label('max_rr')]
        selectable += [
            Activities.calories.label('calories'),
            cls.round_ext_col(Activities, 'avg_temperature'),
            cls.round_ext_col(Activities, 'avg_speed'),
            cls.round_ext_col(Activities, 'max_speed'),
            cls.round_col('avg_step_length'),
        ]
        if include_running_dynamics:
            selectable += [
                cls.round_col('avg_vertical_ratio'),
                cls.avg_gct_balance.label('avg_gct_balance'),
                cls.round_col('avg_vertical_oscillation'),
                cls.avg_ground_contact_time.label('avg_ground_contact_time'),
                cls.avg_stance_time_percent.label('avg_stance_time_percent')
            ]
        selectable += [
            cls.vo2_max.label('vo2_max'),
            Activities.training_effect.label('training_effect'),
            Activities.anaerobic_training_effect.label('anaerobic_training_effect'),
            cls.google_map_loc('start'),
            cls.google_map_loc('stop')
        ]
        return selectable

    @classmethod
    def create_view(cls, db):
        cls._create_activity_view(db, cls._view_selectable(include_sport=True, include_subsport=True, include_type=True, include_course=True))
        cls._create_sport_view(db, cls._view_selectable(), "walking")
        cls._create_sport_view(db, cls._view_selectable(include_course=True, include_subsport=True, include_rr=True, include_running_dynamics=True), "running")
        cls._create_sport_view(db, cls._view_selectable(), "hiking")

    @classmethod
    def create_course_view(cls, db, course_id):
        # The query fails to genarate sql when using the func.round clause.
        selectable = [
            Activities.activity_id.label('activity_id'),
            Activities.name.label('name'),
            Activities.description.label('description'),
            Activities.sport.label('sport'),
            Activities.sub_sport.label('sub_sport'),
            Activities.start_time.label('start_time'),
            Activities.stop_time.label('stop_time'),
            Activities.elapsed_time.label('elapsed_time'),
            cls.round_ext_col(Activities, 'distance'),
            cls.steps.label('steps'),
            cls.avg_pace .label('avg_pace'),
            cls.avg_moving_pace.label('avg_moving_pace'),
            cls.max_pace.label('max_pace'),
            cls.round_col('avg_steps_per_min'),
            cls.round_col('max_steps_per_min'),
            Activities.avg_hr.label('avg_hr'),
            Activities.max_hr.label('max_hr'),
            Activities.avg_rr.label('avg_rr'),
            Activities.max_rr.label('max_rr'),
            Activities.calories.label('calories'),
            cls.round_ext_col(Activities, 'avg_temperature'),
            cls.round_ext_col(Activities, 'avg_speed'),
            cls.round_ext_col(Activities, 'max_speed'),
            cls.round_col('avg_step_length'),
            cls.round_col('avg_vertical_ratio'),
            cls.avg_gct_balance.label('avg_gct_balance'),
            cls.round_col('avg_vertical_oscillation'),
            cls.avg_ground_contact_time.label('avg_ground_contact_time'),
            cls.avg_stance_time_percent.label('avg_stance_time_percent'),
            cls.vo2_max.label('vo2_max'),
            Activities.training_effect.label('training_effect'),
            Activities.anaerobic_training_effect.label('anaerobic_training_effect')
        ]
        cls._create_course_view(db, selectable, course_id)


class PaddleActivities(ActivitiesDb.Base, SportActivities):
    """Paddle based activity table."""

    __tablename__ = 'paddle_activities'

    db = ActivitiesDb
    table_version = 2
    view_version = 5

    strokes = Column(Integer)
    # m or ft
    avg_stroke_distance = Column(Float)

    @classmethod
    def _view_selectable(cls):
        return [
            Activities.activity_id.label('activity_id'),
            Activities.name.label('name'),
            Activities.description.label('description'),
            Activities.sport.label('sport'),
            Activities.sub_sport.label('sub_sport'),
            Activities.start_time.label('start_time'),
            Activities.stop_time.label('stop_time'),
            Activities.elapsed_time.label('elapsed_time'),
            cls.round_ext_col(Activities, 'distance'),
            cls.strokes.label('strokes'),
            cls.round_col('avg_stroke_distance'),
            Activities.avg_cadence.label('avg_cadence'),
            Activities.max_cadence.label('max_cadence'),
            Activities.avg_hr.label('avg_hr'),
            Activities.max_hr.label('max_hr'),
            cls.round_ext_col(Activities, 'calories'),
            cls.round_ext_col(Activities, 'avg_temperature'),
            cls.round_ext_col(Activities, 'avg_speed'),
            cls.round_ext_col(Activities, 'max_speed'),
            Activities.training_effect.label('training_effect'),
            Activities.anaerobic_training_effect.label('anaerobic_training_effect'),
            cls.google_map_loc('start'),
            cls.google_map_loc('stop'),
        ]


class CycleActivities(ActivitiesDb.Base, SportActivities):
    """Cycle based activity table."""

    __tablename__ = 'cycle_activities'

    db = ActivitiesDb
    table_version = 2
    view_version = 6

    strokes = Column(Integer)
    vo2_max = Column(Float)

    @classmethod
    def _view_selectable(cls):
        return [
            Activities.activity_id.label('activity_id'),
            Activities.name.label('name'),
            Activities.description.label('description'),
            Activities.sub_sport.label('sub_sport'),
            Activities.start_time.label('start_time'),
            Activities.stop_time.label('stop_time'),
            Activities.elapsed_time.label('elapsed_time'),
            cls.round_ext_col(Activities, 'distance'),
            cls.strokes.label('strokes'),
            Activities.avg_hr.label('avg_hr'),
            Activities.max_hr.label('max_hr'),
            Activities.avg_rr.label('avg_rr'),
            Activities.max_rr.label('max_rr'),
            Activities.calories.label('calories'),
            cls.round_ext_col(Activities, 'avg_temperature'),
            Activities.avg_cadence.label('avg_rpms'),
            Activities.max_cadence.label('max_rpms'),
            cls.round_ext_col(Activities, 'avg_speed'),
            cls.round_ext_col(Activities, 'max_speed'),
            cls.vo2_max.label('vo2_max'),
            Activities.training_effect.label('training_effect'),
            Activities.anaerobic_training_effect.label('anaerobic_training_effect'),
            cls.google_map_loc('start'),
            cls.google_map_loc('stop'),
        ]



"""A database and database objects for storing health data from a FitBit device."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

# flake8: noqa

from .fitbit_db import FitBitDb, Attributes, DaysSummary
from .import_csv import FitBitData
from .analyze import Analyze



"""Objects for analyzing MS Health data exports."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import sys
import logging
import datetime
import calendar

import fitfile.conversions as conversions

from ..summarydb import SummaryDb, DaysSummary, WeeksSummary, MonthsSummary, YearsSummary
from .mshealth_db import MSHealthDb, MSVaultWeight, DaysSummary as MshDaysSummary


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
stat_logger = logging.getLogger('stats')
stat_logger.addHandler(logging.FileHandler('ms_stats.txt', 'w'))


class Analyze(object):
    """Object for analyzing health data from Microsoft Health."""

    def __init__(self, db_params):
        """Return an instance of the Analyze class."""
        self.mshealthdb = MSHealthDb(db_params)
        self.sumdb = SummaryDb(db_params)

    def __days_from_years(self, year):
        sum_days = DaysSummary.get_days(self.mshealthdb, year)
        weight_days = MSVaultWeight.get_days(self.mshealthdb, year)
        return sum_days + list(set(weight_days) - set(sum_days))

    def __get_days(self, year):
        year_int = int(year)
        days = self.__days_from_years(year)
        days_count = len(days)
        if days_count > 0:
            first_day = days[0]
            last_day = days[-1]
            span = last_day - first_day + 1
        else:
            span = 0
        stat_logger.info("%d Days (%d vs %d): %s", year_int, days_count, span, days)
        for index in range(days_count - 1):
            day = int(days[index])
            next_day = int(days[index + 1])
            if next_day != day + 1:
                day_str = str(conversions.day_of_the_year_to_datetime(year_int, day))
                next_day_str = str(conversions.day_of_the_year_to_datetime(year_int, next_day))
                stat_logger.info("Days gap between %d (%s) and %d (%s)", day, day_str, next_day, next_day_str)

    def __get_months(self, year):
        months = MshDaysSummary.get_month_names(self.mshealthdb, year)
        stat_logger.info("%s Months (%d): %s", year, len(months), months)

    def get_years(self):
        years = MshDaysSummary.get_years(self.mshealthdb)
        stat_logger.info("Years (%d): %s", len(years), years)
        for year in years:
            self.__get_months(year)
            self.__get_days(year)

    def summary(self):
        years = MshDaysSummary.get_years(self.mshealthdb)
        for year in years:
            days = self.__days_from_years(year)
            for day in days:
                day_ts = datetime.date(year, 1, 1) + datetime.timedelta(day - 1)
                stats = MshDaysSummary.get_daily_stats(self.mshealthdb, day_ts)
                stats.update(MSVaultWeight.get_daily_stats(self.mshealthdb, day_ts))
                DaysSummary.insert_or_update(self.sumdb, stats, ignore_none=True)
            for week_starting_day in range(1, 365, 7):
                day_ts = datetime.date(year, 1, 1) + datetime.timedelta(week_starting_day - 1)
                stats = MshDaysSummary.get_weekly_stats(self.mshealthdb, day_ts)
                stats.update(MSVaultWeight.get_weekly_stats(self.mshealthdb, day_ts))
                WeeksSummary.insert_or_update(self.sumdb, stats, ignore_none=True)
            months = MSHealthDb.DaysSummary.get_months(self.mshealthdb, year)
            for month in months:
                start_day_ts = datetime.date(year, month, 1)
                end_day_ts = datetime.date(year, month, calendar.monthrange(year, month)[1])
                stats = MshDaysSummary.get_monthly_stats(self.mshealthdb, start_day_ts, end_day_ts)
                stats.update(MSVaultWeight.get_monthly_stats(self.mshealthdb, start_day_ts, end_day_ts))
                MonthsSummary.insert_or_update(self.sumdb, stats, ignore_none=True)
            stats.update(MSVaultWeight.get_yearly_stats(self.mshealthdb, year))
            YearsSummary.insert_or_update(self.sumdb, stats, ignore_none=True)



"""Garmin Database reading and writing library."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

# flake8: noqa

from .garmin_db import GarminDb, Attributes, Device, DeviceInfo, File, Weight, Stress, Sleep, SleepEvents, RestingHeartRate, DailySummary
from .monitoring_db import MonitoringDb, MonitoringInfo, MonitoringHeartRate, MonitoringIntensity, MonitoringClimb, Monitoring, \
    MonitoringRespirationRate, MonitoringPulseOx
from .activities_db import ActivitiesDb, ActivitiesLocationSegment, Activities, ActivityLaps, ActivityRecords, SportActivities, StepsActivities, \
    PaddleActivities, CycleActivities
from .garmin_summary_db import GarminSummaryDb, Summary, YearsSummary, MonthsSummary, WeeksSummary, DaysSummary, IntensityHR



"""A database and database objects for storing health data from Garmin Connect."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

# flake8: noqa

from .version_info import version_string

__version__ = version_string()

from .activity_fit_file_processor import ActivityFitFileProcessor
from .fit_data import FitData
from .fit_file_processor import FitFileProcessor
from .garmin_connect_config_manager import GarminConnectConfigManager
from .config_manager import ConfigManager
from .graphs import Graph
from .statistics import Statistics
from .tcx import Tcx
from .monitoring_fit_file_processor import MonitoringFitFileProcessor
from .export_activities import ActivityExporter
from .open_with_basecamp import OpenWithBaseCamp
from .open_with_google_earth import OpenWithGoogleEarth

from .copy import Copy
from .download import Download
from .analyze import Analyze
from .plugin_manager import PluginManager
from .version import format_version, log_version, python_version_check

from .import_monitoring import GarminMonitoringFitData, GarminSummaryData, GarminProfile, GarminWeightData, GarminSleepData, GarminRhrData, GarminSettingsFitData, GarminHydrationData
from .import_activities import GarminActivitiesFitData, GarminTcxData, GarminJsonSummaryData, GarminJsonDetailsData



"""Objects representing a database and database objects for storing health data from a FitBit device."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging
import datetime
from sqlalchemy import Column, Integer, Date, Float, Time

import fitfile
import idbutils


logger = logging.getLogger(__name__)

FitBitDb = idbutils.DB.create('fitbit', 2, "Database for storing health data from FitBit.")
Attributes = idbutils.DbObject.create('attributes', FitBitDb, 1, base=idbutils.KeyValueObject, doc="key-value data from a FitBit device.")


class DaysSummary(FitBitDb.Base, idbutils.DbObject):
    """A table that holds summarized information about a day with one row per day."""

    __tablename__ = 'days_summary'

    db = FitBitDb
    table_version = 1

    day = Column(Date, primary_key=True)
    calories_in = Column(Integer)
    log_water = Column(Float)
    calories = Column(Integer)
    calories_bmr = Column(Integer)
    steps = Column(Integer)
    distance = Column(Float)
    floors = Column(Integer)
    elevation = Column(Float)
    sedentary_mins = Column(Integer)
    lightly_active_mins = Column(Integer)
    fairly_active_mins = Column(Integer)
    very_active_mins = Column(Integer)
    activities_calories = Column(Integer)
    sleep_start = Column(Time)
    in_bed_mins = Column(Integer)
    asleep_mins = Column(Integer)
    awakenings_count = Column(Integer)
    awake_mins = Column(Integer)
    to_fall_asleep_mins = Column(Integer)
    after_wakeup_mins = Column(Integer)
    sleep_efficiency = Column(Integer)
    weight = Column(Float)
    bmi = Column(Float)

    @classmethod
    def __get_activity_mins_stats(cls, db, func, start_ts, end_ts):
        moderate_activity_time = fitfile.conversions.min_to_dt_time(func(db, cls.fairly_active_mins, start_ts, end_ts))
        vigorous_activity_time = fitfile.conversions.min_to_dt_time(func(db, cls.very_active_mins, start_ts, end_ts))
        intensity_time = datetime.time.min
        if moderate_activity_time:
            intensity_time = fitfile.conversions.add_time(intensity_time, moderate_activity_time)
        if vigorous_activity_time:
            intensity_time = fitfile.conversions.add_time(intensity_time, vigorous_activity_time, 2)
        stats = {
            'intensity_time': intensity_time,
            'moderate_activity_time': moderate_activity_time,
            'vigorous_activity_time': vigorous_activity_time,
        }
        return stats

    @classmethod
    def __get_floors_stats(cls, db, func, start_ts, end_ts):
        return {'floors': func(db, cls.floors, start_ts, end_ts)}

    @classmethod
    def __get_steps_stats(cls, db, func, start_ts, end_ts):
        return {'steps': func(db, cls.steps, start_ts, end_ts)}

    @classmethod
    def __get_weight_stats(cls, db, start_ts, end_ts):
        stats = {
            'weight_avg': cls.get_col_avg(db, cls.weight, start_ts, end_ts, True),
            'weight_min': cls.get_col_min(db, cls.weight, start_ts, end_ts, True),
            'weight_max': cls.get_col_max(db, cls.weight, start_ts, end_ts),
        }
        return stats

    @classmethod
    def __get_sleep_stats(cls, db, start_ts, end_ts):
        return {
            'sleep_avg': fitfile.conversions.min_to_dt_time(cls.get_col_avg(db, cls.asleep_mins, start_ts, end_ts, True)),
            'sleep_min': fitfile.conversions.min_to_dt_time(cls.get_col_min(db, cls.asleep_mins, start_ts, end_ts, True)),
            'sleep_max': fitfile.conversions.min_to_dt_time(cls.get_col_max(db, cls.asleep_mins, start_ts, end_ts)),
        }

    @classmethod
    def __get_calories_stats(cls, db, start_ts, end_ts):
        calories_bmr_avg = cls.get_col_avg(db, cls.calories_bmr, start_ts, end_ts)
        calories_active_avg = cls.get_col_avg(db, cls.activities_calories, start_ts, end_ts)
        if calories_bmr_avg is not None and calories_active_avg is not None:
            calories_avg = calories_bmr_avg + calories_active_avg
        else:
            calories_avg = None
        return {
            'calories_avg': calories_avg,
            'calories_bmr_avg': calories_bmr_avg,
            'calories_active_avg': calories_active_avg,
        }

    @classmethod
    def get_daily_stats(cls, db, day_ts):
        """Return a dictionary of aggregate statistics for the given day."""
        stats = cls.__get_activity_mins_stats(db, cls.get_col_sum, day_ts, day_ts + datetime.timedelta(1))
        stats.update(cls.__get_floors_stats(db, cls.get_col_sum, day_ts, day_ts + datetime.timedelta(1)))
        stats.update(cls.__get_steps_stats(db, cls.get_col_sum, day_ts, day_ts + datetime.timedelta(1)))
        stats.update(cls.__get_weight_stats(db, day_ts, day_ts + datetime.timedelta(1)))
        stats.update(cls.__get_sleep_stats(db, day_ts, day_ts + datetime.timedelta(1)))
        stats.update(cls.__get_calories_stats(db, day_ts, day_ts + datetime.timedelta(1)))
        stats['day'] = day_ts
        return stats

    @classmethod
    def get_weekly_stats(cls, db, first_day_ts):
        """Return a dictionary of aggregate statistics for the given week."""
        stats = cls.__get_activity_mins_stats(db, cls.get_col_sum, first_day_ts, first_day_ts + datetime.timedelta(7))
        stats.update(cls.__get_floors_stats(db, cls.get_col_sum, first_day_ts, first_day_ts + datetime.timedelta(7)))
        stats.update(cls.__get_steps_stats(db, cls.get_col_sum, first_day_ts, first_day_ts + datetime.timedelta(7)))
        stats.update(cls.__get_weight_stats(db, first_day_ts, first_day_ts + datetime.timedelta(7)))
        stats.update(cls.__get_sleep_stats(db, first_day_ts, first_day_ts + datetime.timedelta(7)))
        stats.update(cls.__get_calories_stats(db, first_day_ts, first_day_ts + datetime.timedelta(7)))
        stats['first_day'] = first_day_ts
        return stats

    @classmethod
    def get_monthly_stats(cls, db, first_day_ts, last_day_ts):
        """Return a dictionary of aggregate statistics for the given month."""
        stats = cls.__get_activity_mins_stats(db, cls.get_col_sum, first_day_ts, last_day_ts)
        stats.update(cls.__get_floors_stats(db, cls.get_col_sum, first_day_ts, last_day_ts))
        stats.update(cls.__get_steps_stats(db, cls.get_col_sum, first_day_ts, last_day_ts))
        stats.update(cls.__get_weight_stats(db, first_day_ts, last_day_ts))
        stats.update(cls.__get_sleep_stats(db, first_day_ts, last_day_ts))
        stats.update(cls.__get_calories_stats(db, first_day_ts, last_day_ts))
        stats['first_day'] = first_day_ts
        return stats

    @classmethod
    def get_yearly_stats(cls, db, year):
        """Return a dictionary of aggregate statistics for the given year."""
        first_day_ts = datetime.datetime(year, 1, 1)
        last_day_ts = first_day_ts + datetime.timedelta(365)
        stats = cls.__get_activity_mins_stats(db, cls.get_col_sum, first_day_ts, last_day_ts)
        stats.update(cls.__get_floors_stats(db, cls.get_col_sum, first_day_ts, last_day_ts))
        stats.update(cls.__get_steps_stats(db, cls.get_col_sum, first_day_ts, last_day_ts))
        stats.update(cls.__get_weight_stats(db, first_day_ts, last_day_ts))
        stats.update(cls.__get_sleep_stats(db, first_day_ts, last_day_ts))
        stats.update(cls.__get_calories_stats(db, first_day_ts, last_day_ts))
        stats['first_day'] = first_day_ts
        return stats



#!/usr/bin/env python3

"""Script for importing into a DB and summarizing CSV formatted FitBit export data."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import sys
import argparse
import logging


from garmindb.fitbitdb import FitBitDb, FitBitData, Analyze
from garmindb import ConfigManager
from garmindb.version import format_version


logging.basicConfig(filename='fitbit.log', filemode='w', level=logging.DEBUG)
logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


def usage(program):
    """Print the usage info for the script."""
    print('%s -i <inputfile> ...' % program)
    sys.exit()


def main(argv):
    """Import into a DB and summarize CSV formatted FitBit export data."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", help="print the program's version", action='version', version=format_version(sys.argv[0]))
    parser.add_argument("-t", "--trace", help="Turn on debug tracing", type=int, default=0)
    modes_group = parser.add_argument_group('Modes')
    modes_group.add_argument("-i", "--input_file", help="Specifiy the CSV file to import into the database")
    modes_group.add_argument("--delete_db", help="Delete FitBit db file.", action="store_true", default=False)
    args = parser.parse_args()

    root_logger = logging.getLogger()
    if args.trace > 0:
        root_logger.setLevel(logging.DEBUG)
    else:
        root_logger.setLevel(logging.INFO)

    db_params = ConfigManager.get_db_params()

    if args.delete_db:
        FitBitDb.delete_db(db_params)
        sys.exit()

    fitbit_dir = ConfigManager.get_or_create_fitbit_dir()
    metric = ConfigManager.get_metric()
    fd = FitBitData(args.input_file, fitbit_dir, db_params, metric, args.trace)
    if fd.file_count() > 0:
        fd.process_files()

    analyze = Analyze(db_params)
    analyze.get_years()
    analyze.summary()


if __name__ == "__main__":
    main(sys.argv[1:])



"""Test garmin db."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging
import datetime

import fitfile

from garmindb import ConfigManager
from garmindb.garmindb import GarminDb, Attributes, Device, DeviceInfo, File, Weight, Stress, Sleep, SleepEvents, RestingHeartRate

from test_db_base import TestDBBase


root_logger = logging.getLogger()
handler = logging.FileHandler('garmin_db.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestGarminDb(TestDBBase, unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        db_params = ConfigManager.get_db_params()
        cls.garmin_db = GarminDb(db_params)
        table_dict = {
            'attributes_table': Attributes,
            'device_table': Device,
            'device_info_table': DeviceInfo,
            'file_table': File,
            'weight_table': Weight,
            'stress_table': Stress,
            'sleep_table': Sleep,
            'sleep_events_table': SleepEvents,
            'resting_heart_rate_table': RestingHeartRate
        }
        super().setUpClass(cls.garmin_db, table_dict)

    def check_col_stat(self, value_name, value, bounds):
        min_value, max_value = bounds
        self.assertGreaterEqual(value, min_value, '%s value %s less than min %s' % (value_name, value, min_value))
        self.assertLessEqual(value, max_value, '%s value %s greater than max %s' % (value_name, value, max_value))
        logger.info("%s: %s", value_name, value)

    def check_col_stats(self, db, table, col, col_name, ignore_le_zero, time_col,
                        records_bounds, max_bounds, min_bounds, avg_bounds, latest_bounds):
        self.check_col_stat(col_name + ' records', table.row_count(db), records_bounds)
        if time_col:
            maximum = table.get_time_col_max(db, col)
        else:
            maximum = table.get_col_max(db, col)
        self.check_col_stat(col_name + ' max', maximum, max_bounds)
        if time_col:
            minimum = table.get_time_col_min(db, col, None, None)
        else:
            minimum = table.get_col_min(db, col, None, None, ignore_le_zero)
        self.check_col_stat(col_name + ' min', minimum, min_bounds)
        if time_col:
            average = table.get_time_col_avg(db, col, None, None)
        else:
            average = table.get_col_avg(db, col, None, None, ignore_le_zero)
        self.check_col_stat(col_name + ' avg', average, avg_bounds)
        latest = table.get_time_col_latest(db, col) if time_col else table.get_col_latest(db, col, ignore_le_zero)
        self.check_col_stat(col_name + ' latest', latest, latest_bounds)

    def test_garmindb_tables_bounds(self):
        self.check_col_stats(
            self.garmin_db, Weight, Weight.weight, 'Weight', False, False,
            (0, 10 * 365),
            (25, 300),
            (25, 300),
            (25, 300),
            (25, 300)
        )
        stress_min = -2
        stress_max = 100
        self.check_col_stats(
            self.garmin_db, Stress, Stress.stress, 'Stress', True, False,
            (1, 10000000),
            (25, 100),
            (stress_min, 2),
            (stress_min, stress_max),
            (stress_min, stress_max)
        )
        self.check_col_stats(
            self.garmin_db, RestingHeartRate, RestingHeartRate.resting_heart_rate, 'RHR', True, False,
            (1, 10000000),
            (30, 100),
            (30, 100),
            (30, 100),
            (30, 100)
        )
        self.check_col_stats(
            self.garmin_db, Sleep, Sleep.total_sleep, 'Sleep', True, True,
            (1, 10000000),
            (datetime.time(8), datetime.time(16)),
            (datetime.time(0), datetime.time(4)),
            (datetime.time(4), datetime.time(10)),
            (datetime.time(2), datetime.time(16))
        )
        self.check_col_stats(
            self.garmin_db, Sleep, Sleep.rem_sleep, 'REM Sleep', True, True,
            (1, 10000000),
            (datetime.time(2), datetime.time(8)),           # max
            (datetime.time(0), datetime.time(2)),           # min
            (datetime.time(minute=30), datetime.time(6)),   # avg
            (datetime.time(minute=10), datetime.time(6))    # latest
        )

    def test_measurement_system(self):
        measurement_system = Attributes.measurements_type(self.garmin_db, fitfile.field_enums.DisplayMeasure.metric)
        self.assertIn(measurement_system, fitfile.field_enums.DisplayMeasure)


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3

"""A script that generated graphs for health data in a database."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging
import sys
import datetime
import enum

import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from .garmindb import MonitoringDb, Monitoring, MonitoringHeartRate
from .summarydb import DaysSummary, WeeksSummary, MonthsSummary, SummaryDb
from .config_manager import ConfigManager
from .garmin_connect_config_manager import GarminConnectConfigManager


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()

gc_config = GarminConnectConfigManager()


class YAxisLabelPostion(enum.Enum):
    """An enum of the label postions for the Y axis."""

    right   = 0
    left    = 1

    @classmethod
    def from_integer(cls, integer):
        """Create an instance of a YAxisLabelPostion enum from an integer."""
        return YAxisLabelPostion(integer % 2)


class Colors(enum.Enum):
    """An enum of the colors used for generating graphs."""

    b   = 0
    g   = 1
    r   = 2
    c   = 3
    m   = 4
    y   = 5
    k   = 6
    w   = 7

    @classmethod
    def from_integer(cls, integer):
        """Create an instance of a Color enum from an integer."""
        return Colors(integer % 8)


class Graph(object):
    """A class that generates graphs for GarminDb data sets."""

    __table = {
        'days'      : DaysSummary,
        'weeks'     : WeeksSummary,
        'months'    : MonthsSummary
    }

    def __init__(self, debug=False, save=False):
        """Return an instance of the Graph class."""
        self.debug = debug
        self.save = save

    @classmethod
    def __remove_discontinuities(cls, data):
        last = 0
        for index, entry in enumerate(data):
            if not entry:
                data[index] = last
            else:
                last = data[index]
        return data

    @classmethod
    def __graph_multiple_single_axes(cls, time, data_list, stat_name, ylabel, save, geometry=111):
        title = f'{stat_name} Over Time'
        figure = plt.figure(figsize=ConfigManager.get_graphs('size'))
        for index, data in enumerate(data_list):
            color = Colors.from_integer(index).name
            axes = figure.add_subplot(geometry, frame_on=(index == 0))
            axes.plot(time, data, color=color)
            axes.grid()
        axes.set_title(title)
        axes.set_xlabel('Time')
        axes.set_ylabel(ylabel)
        if save:
            figure.savefig(stat_name + ".png")

    @classmethod
    def __graph_multiple(cls, time, data_list, stat_name, period, ylabel_list, yrange_list, save, geometry=111):
        title = f'{stat_name} by {period}'
        figure = plt.figure(figsize=ConfigManager.get_graphs('size'))
        for index, data in enumerate(data_list):
            color = Colors.from_integer(index).name
            axes = figure.add_subplot(geometry, label=ylabel_list[index], frame_on=(index == 0))
            axes.plot(time, data, color=color)
            axes.set_ylabel(ylabel_list[index], color=color)
            axes.yaxis.set_label_position(YAxisLabelPostion.from_integer(index).name)
            if (index % 2) == 0:
                axes.yaxis.tick_right()
            else:
                axes.yaxis.tick_left()
            axes.tick_params(axis='y', colors=color)
            if yrange_list is None:
                axes.set_ylim([min(data), max(data)])
            else:
                axes.set_ylim(yrange_list[index])
            axes.grid()
        axes.set_title(title)
        axes.set_xlabel('Time')
        if save:
            figure.savefig(stat_name + ".png")

    @classmethod
    def __graph_over(cls, date, over_data_dicts, under_data_dict, title, xlabel, ylabel, save_name=None, geometry=111):
        figure = plt.figure(figsize=ConfigManager.get_graphs('size'))
        # First graph the data that appears under
        axes = figure.add_subplot(geometry, frame_on=True)
        axes.fill_between(under_data_dict['time'], under_data_dict['data'], 0, color=Colors.c.name)
        axes.set_ylim(under_data_dict['limits'])
        axes.set_xticks([])
        axes.set_yticks([])
        # then graph the data that appears on top
        colors = [Colors.r.name, Colors.b.name]
        for index, _ in enumerate(over_data_dicts):
            over_data_dict = over_data_dicts[index]
            color = colors[index]
            label = over_data_dict['label']
            axes = figure.add_subplot(geometry, frame_on=False, label=label)
            axes.plot(over_data_dict['time'], over_data_dict['data'], color=color)
            axes.set_ylabel(label, color=color)
            axes.yaxis.set_label_position(YAxisLabelPostion.from_integer(index).name)
            if (index % 2) == 0:
                axes.yaxis.tick_right()
                axes.set_xticks([])
            else:
                axes.yaxis.tick_left()
            limits = over_data_dicts[index].get('limits')
            if limits is not None:
                axes.set_ylim(limits)
            axes.grid()
        axes.set_title(title)
        axes.set_xlabel(xlabel)
        x_format = mdates.DateFormatter('%H:%M')
        axes.xaxis.set_major_formatter(x_format)
        if save_name:
            figure.savefig(save_name)

    def _graph_steps(self, time, data, period, geometry=111):
        steps = self.__remove_discontinuities([entry.steps for entry in data])
        steps_goal_percent = self.__remove_discontinuities([entry.steps_goal_percent for entry in data])
        yrange_list = [(0, max(steps) * 1.1), (0, max(steps_goal_percent) * 2)]
        self.__graph_multiple(time, [steps, steps_goal_percent], 'Steps', period, ['Steps', 'Step Goal Percent'], yrange_list, self.save, geometry)

    def _graph_hr(self, time, data, period, geometry=111):
        rhr = self.__remove_discontinuities([entry.rhr_avg for entry in data])
        inactive_hr = self.__remove_discontinuities([entry.inactive_hr_avg for entry in data])
        self.__graph_multiple(time, [rhr, inactive_hr], 'Heart Rate', period, ['RHR', 'Inactive hr'], [(30, 100), (30, 100)], self.save, geometry)

    def _graph_itime(self, time, data, period, geometry=111):
        itime = [entry.intensity_time_mins for entry in data]
        itime_goal_percent = self.__remove_discontinuities([entry.intensity_time_goal_percent for entry in data])
        itime_goal_max = max([entry.intensity_time_goal_mins for entry in data])
        yrange_list = [(0, itime_goal_max * 5), (0, max(itime_goal_percent) * 1.1)]
        self.__graph_multiple(time, [itime, itime_goal_percent], 'Intensity Minutes', period, ['Intensity Minutes', 'Intensity Minutes Goal Percent'],
                              yrange_list, self.save, geometry)

    def _graph_weight(self, time, data, period, geometry=111):
        weight = [entry.weight_avg for entry in data]
        self.__graph_multiple_single_axes(time, [weight], 'Weight', 'weight', self.save, geometry)

    def graph_activity(self, activity, period=None, days=None, geometry=111):
        """Generate a graph for the given activity with points every period spanning days."""
        if period is None:
            period = ConfigManager.graphs_activity_config(activity, 'period')
        if days is None:
            days = ConfigManager.graphs_activity_config(activity, 'days')
        db_params = ConfigManager.get_db_params()
        sum_db = SummaryDb(db_params, self.debug)
        end_ts = datetime.datetime.now()
        start_ts = end_ts - datetime.timedelta(days=days)
        table = self.__table[period]
        data = table.get_for_period(sum_db, start_ts, end_ts, table)
        if period == 'days':
            time = [entry.day for entry in data]
        else:
            time = [entry.first_day for entry in data]
        graph_func_name = '_graph_' + activity
        graph_func = getattr(self, graph_func_name, None)
        graph_func(time, data, period, geometry)

    def __format_steps(self, data):
        steps = []
        steps_count = {}
        for entry in data:
            if entry.steps is not None:
                if entry.activity_type in steps_count:
                    if entry.steps > steps_count[entry.activity_type]:
                        steps_count[entry.activity_type] = entry.steps
                else:
                    steps_count[entry.activity_type] = entry.steps
            steps.append(sum(steps_count.values()))
        return steps

    def graph_date(self, date, geometry=111):
        """Generate a graph for the given date."""
        if date is None:
            date = (datetime.datetime.now() - datetime.timedelta(days=1)).date()
        db_params = ConfigManager.get_db_params()
        mon_db = MonitoringDb(db_params, self.debug)
        start_ts = datetime.datetime.combine(date, datetime.datetime.min.time())
        end_ts = datetime.datetime.combine(date, datetime.datetime.max.time())
        hr_data = MonitoringHeartRate.get_for_period(mon_db, start_ts, end_ts, MonitoringHeartRate)
        data = Monitoring.get_for_period(mon_db, start_ts, end_ts, Monitoring)
        over_data_dict = [
            {
                'label'     : 'Cumulative Steps',
                'time'      : [entry.timestamp for entry in data],
                'data'      : self.__format_steps(data),
            },
            {
                'label'     : 'Heart Rate',
                'time'      : [entry.timestamp for entry in hr_data],
                'data'      : [entry.heart_rate for entry in hr_data],
                'limits'    : (30, 220)
            }
        ]
        under_data_dict = {
            'time'      : [entry.timestamp for entry in data],
            'data'      : self.__remove_discontinuities([entry.intensity for entry in data]),
            'limits'    : (0, 10)
        }
        save_name = f"{date}_daily.png" if self.save else None
        self.__graph_over(date, over_data_dict, under_data_dict, f'Daily Summary for {date}: Heart Rate and Steps over Activity',
                          'Time of Day', 'heart rate', save_name=save_name, geometry=geometry)



"""Class that takes a parsed activity FIT file object and imports it into a database."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging
import sys

import fitfile

from .garmindb import File, ActivitiesDb, Activities, ActivityRecords, ActivityLaps, StepsActivities, CycleActivities, PaddleActivities
from .fit_file_processor import FitFileProcessor


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class ActivityFitFileProcessor(FitFileProcessor):
    """Class that takes a parsed activity FIT file object and imports it into a database."""

    def write_file(self, fit_file):
        """Given a Fit File object, write all of its messages to the DB."""
        self.activity_file_plugins = [plugin for plugin in self.plugin_manager.get_activity_file_processors(fit_file).values()]
        if len(self.activity_file_plugins):
            root_logger.info("Loaded %d activity plugins %r for file %s", len(self.activity_file_plugins), self.activity_file_plugins, fit_file)
        # Create the db after setting up the plugins so that plugin tables are handled properly
        self.garmin_act_db = ActivitiesDb(self.db_params, self.debug - 1)
        with self.garmin_db.managed_session() as self.garmin_db_session, self.garmin_act_db.managed_session() as self.garmin_act_db_session:
            self._write_message_types(fit_file, fit_file.message_types)

    def _plugin_dispatch(self, handler_name, *args, **kwargs):
        result = {}
        for plugin in self.activity_file_plugins:
            function = getattr(plugin, handler_name, None)
            if function:
                result.update(function(*args, **kwargs))
        return result

    def _write_lap(self, fit_file, message_type, messages):
        """Write all lap messages to the database."""
        for lap_num, message in enumerate(messages):
            self._write_lap_entry(fit_file, message.fields, lap_num)

    def _write_record(self, fit_file, message_type, messages):
        """Write all record messages to the database."""
        for record_num, message in enumerate(messages):
            self._write_record_entry(fit_file, message.fields, record_num)

    def _write_record_entry(self, fit_file, message_fields, record_num):
        # We don't get record data from multiple sources so we don't need to coellesce data in the DB.
        # It's fastest to just write the new data out if it doesn't currently exist.
        activity_id = File.id_from_path(fit_file.filename)
        plugin_record = self._plugin_dispatch('write_record_entry', self.garmin_act_db_session, fit_file, activity_id, message_fields, record_num)
        if not ActivityRecords.s_exists(self.garmin_act_db_session, {'activity_id' : activity_id, 'record' : record_num}):
            record = {
                'activity_id'                       : activity_id,
                'record'                            : record_num,
                'timestamp'                         : fit_file.utc_datetime_to_local(message_fields.timestamp),
                'position_lat'                      : message_fields.get('position_lat'),
                'position_long'                     : message_fields.get('position_long'),
                'distance'                          : message_fields.get('distance'),
                'cadence'                           : message_fields.get('cadence'),
                'hr'                                : message_fields.get('heart_rate'),
                'rr'                                : message_fields.get('respiration_rate'),
                'altitude'                          : message_fields.get('altitude'),
                'speed'                             : message_fields.get('speed'),
                'temperature'                       : message_fields.get('temperature'),
            }
            record.update(plugin_record)
            root_logger.debug("_write_record_entry activity_id %s, record %s doesn't exist", activity_id, record_num)
            self.garmin_act_db_session.add(ActivityRecords(**record))

    def _write_lap_entry(self, fit_file, message_fields, lap_num):
        # we don't get laps data from multiple sources so we don't need to coellesce data in the DB.
        # It's fastest to just write new data out if the it doesn't currently exist.
        activity_id = File.id_from_path(fit_file.filename)
        plugin_lap = self._plugin_dispatch('write_lap_entry', self.garmin_act_db_session, fit_file, activity_id, message_fields, lap_num)
        if not ActivityLaps.s_exists(self.garmin_act_db_session, {'activity_id' : activity_id, 'lap' : lap_num}):
            lap = {
                'activity_id'                       : File.id_from_path(fit_file.filename),
                'lap'                               : lap_num,
                'start_time'                        : fit_file.utc_datetime_to_local(message_fields.start_time),
                'stop_time'                         : fit_file.utc_datetime_to_local(message_fields.timestamp),
                'elapsed_time'                      : message_fields.get('total_elapsed_time'),
                'moving_time'                       : message_fields.get('total_timer_time'),
                'start_lat'                         : message_fields.get('start_position_lat'),
                'start_long'                        : message_fields.get('start_position_long'),
                'stop_lat'                          : message_fields.get('end_position_lat'),
                'stop_long'                         : message_fields.get('end_position_long'),
                'distance'                          : message_fields.get('total_distance'),
                'cycles'                            : message_fields.get('total_cycles'),
                'avg_hr'                            : message_fields.get('avg_heart_rate'),
                'max_hr'                            : message_fields.get('max_heart_rate'),
                'avg_rr'                            : message_fields.get('avg_respiration_rate'),
                'max_rr'                            : message_fields.get('max_respiration_rate'),
                'calories'                          : message_fields.get('total_calories'),
                'avg_cadence'                       : message_fields.get('avg_cadence'),
                'max_cadence'                       : message_fields.get('max_cadence'),
                'avg_speed'                         : message_fields.get('avg_speed'),
                'max_speed'                         : message_fields.get('max_speed'),
                'ascent'                            : message_fields.get('total_ascent'),
                'descent'                           : message_fields.get('total_descent'),
                'max_temperature'                   : message_fields.get('max_temperature'),
                'avg_temperature'                   : message_fields.get('avg_temperature'),
            }
            lap.update(plugin_lap)
            root_logger.debug("writing lap %r for %s", lap, fit_file.filename)
            self.garmin_act_db_session.add(ActivityLaps(**lap))

    def _write_steps_entry(self, fit_file, activity_id, sub_sport, message_fields):
        steps = {
            'activity_id'                       : activity_id,
            'steps'                             : message_fields.get('total_steps'),
            'avg_pace'                          : fitfile.conversions.perhour_speed_to_pace(message_fields.avg_speed),
            'max_pace'                          : fitfile.conversions.perhour_speed_to_pace(message_fields.max_speed),
            'avg_steps_per_min'                 : message_fields.get('avg_steps_per_min'),
            'max_steps_per_min'                 : message_fields.get('max_steps_per_min'),
            'avg_step_length'                   : message_fields.get('avg_step_length'),
            'avg_vertical_ratio'                : message_fields.get('avg_vertical_ratio'),
            'avg_vertical_oscillation'          : message_fields.get('avg_vertical_oscillation'),
            'avg_gct_balance'                   : message_fields.get('avg_stance_time_balance'),
            'avg_ground_contact_time'           : message_fields.get('avg_stance_time'),
            'avg_stance_time_percent'           : message_fields.get('avg_stance_time_percent'),
        }
        steps.update(self._plugin_dispatch('write_steps_entry', self.garmin_act_db_session, fit_file, activity_id, sub_sport, message_fields))
        root_logger.debug("_write_steps_entry: %r", steps)
        StepsActivities.s_insert_or_update(self.garmin_act_db_session, steps, ignore_none=True, ignore_zero=True)

    def _write_running_entry(self, fit_file, activity_id, sub_sport, message_fields):
        return self._write_steps_entry(fit_file, activity_id, sub_sport, message_fields)

    def _write_walking_entry(self, fit_file, activity_id, sub_sport, message_fields):
        return self._write_steps_entry(fit_file, activity_id, sub_sport, message_fields)

    def _write_hiking_entry(self, fit_file, activity_id, sub_sport, message_fields):
        return self._write_steps_entry(fit_file, activity_id, sub_sport, message_fields)

    def _write_cycling_entry(self, fit_file, activity_id, sub_sport, message_fields):
        ride = {
            'activity_id'   : activity_id,
            'strokes'       : message_fields.get('total_strokes'),
        }
        ride.update(self._plugin_dispatch('write_cycle_entry', self.garmin_act_db_session, fit_file, activity_id, sub_sport, message_fields))
        CycleActivities.s_insert_or_update(self.garmin_act_db_session, ride, ignore_none=True, ignore_zero=True)

    def _write_stand_up_paddleboarding_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("sup sport entry: %r", message_fields)
        paddle = {
            'activity_id'           : activity_id,
            'strokes'               : message_fields.get('total_strokes'),
            'avg_stroke_distance'   : message_fields.get('avg_stroke_distance'),
        }
        paddle.update(self._plugin_dispatch('write_paddle_entry', self.garmin_act_db_session, fit_file, activity_id, sub_sport, message_fields))
        PaddleActivities.s_insert_or_update(self.garmin_act_db_session, paddle, ignore_none=True, ignore_zero=True)

    def _write_rowing_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("row sport entry: %r", message_fields)
        return self._write_stand_up_paddleboarding_entry(fit_file, activity_id, sub_sport, message_fields)

    def _write_boating_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("boating sport entry: %r", message_fields)

    def _write_fitness_equipment_entry(self, fit_file, activity_id, sub_sport, message_fields):
        try:
            function = getattr(self, '_write_' + sub_sport.name + '_entry')
            function(fit_file, activity_id, sub_sport, message_fields)
        except AttributeError:
            root_logger.info("No sub sport handler type %s from %s: %s", sub_sport, fit_file.filename, message_fields)

    def _write_alpine_skiing_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("Skiing sport entry: %r", message_fields)

    def _write_swimming_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("Swimming sport entry: %r", message_fields)

    def _write_training_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("Training sport entry: %r", message_fields)

    def _write_transition_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("Transition sport entry: %r", message_fields)

    def _write_generic_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("Generic sport entry: %r", message_fields)

    def __choose_sport(self, current_sport, current_sub_sport, new_sport, new_sub_sport):
        sport = fitfile.Sport.strict_from_string(current_sport)
        sub_sport = fitfile.SubSport.strict_from_string(current_sub_sport)
        if new_sport is not None and (sport is None or (not sport.preferred() and new_sport.preferred())):
            sport = new_sport
        if new_sub_sport is not None and (sub_sport is None or (not sub_sport.preferred() and new_sub_sport.preferred())):
            sub_sport = new_sub_sport
        return {'sport' : fitfile.field_enums.name_for_enum(sport), 'sub_sport' : fitfile.field_enums.name_for_enum(sub_sport)}

    def _write_session_entry(self, fit_file, message_fields):
        activity_id = File.id_from_path(fit_file.filename)
        sport = message_fields.sport
        sub_sport = message_fields.sub_sport
        activity = {
            'activity_id'                       : activity_id,
            'start_time'                        : fit_file.utc_datetime_to_local(message_fields.start_time),
            'stop_time'                         : fit_file.utc_datetime_to_local(message_fields.timestamp),
            'elapsed_time'                      : message_fields.total_elapsed_time,
            'moving_time'                       : message_fields.get('total_timer_time'),
            'start_lat'                         : message_fields.get('start_position_lat'),
            'start_long'                        : message_fields.get('start_position_long'),
            'stop_lat'                          : message_fields.get('end_position_lat'),
            'stop_long'                         : message_fields.get('end_position_long'),
            'distance'                          : message_fields.get('total_distance'),
            'cycles'                            : message_fields.get('total_cycles'),
            'laps'                              : message_fields.get('num_laps'),
            'avg_hr'                            : message_fields.get('avg_heart_rate'),
            'max_hr'                            : message_fields.get('max_heart_rate'),
            'avg_rr'                            : message_fields.get('avg_respiration_rate'),
            'max_rr'                            : message_fields.get('max_respiration_rate'),
            'calories'                          : message_fields.get('total_calories'),
            'avg_cadence'                       : message_fields.get('avg_cadence'),
            'max_cadence'                       : message_fields.get('max_cadence'),
            'avg_speed'                         : message_fields.get('avg_speed'),
            'max_speed'                         : message_fields.get('max_speed'),
            'ascent'                            : message_fields.get('total_ascent'),
            'descent'                           : message_fields.get('total_descent'),
            'max_temperature'                   : message_fields.get('max_temperature'),
            'avg_temperature'                   : message_fields.get('avg_temperature'),
            'training_effect'                   : message_fields.get('total_training_effect'),
            'anaerobic_training_effect'         : message_fields.get('total_anaerobic_training_effect')
        }
        activity.update(self._plugin_dispatch('write_session_entry', self.garmin_act_db_session, fit_file, activity_id, message_fields))
        # json metadata gives better values for sport and subsport, so use existing value if set
        current = Activities.s_get(self.garmin_act_db_session, activity_id)
        if current:
            activity.update(self.__choose_sport(current.sport, current.sub_sport, sport, sub_sport))
            root_logger.debug("Updating with %r", activity)
            current.update_from_dict(activity, ignore_none=True, ignore_zero=True)
        else:
            activity.update({'sport': sport.name, 'sub_sport': sub_sport.name})
            root_logger.debug("Adding %r", activity)
            self.garmin_act_db_session.add(Activities(**activity))
        if sport is not None:
            function_name = '_write_' + sport.name + '_entry'
            try:
                function = getattr(self, function_name, None)
                if function is not None:
                    function(fit_file, activity_id, sub_sport, message_fields)
                else:
                    root_logger.warning("No sport handler for type %s from %s: %s", sport, fit_file.filename, message_fields)
            except Exception as e:
                root_logger.error("Exception in %s from %s: %s", function_name, fit_file.filename, e)



"""Objects for exporting Garmin activity data."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import os

from fitfile import Distance, Speed

from .garmindb import GarminDb, File, Device, ActivitiesDb, Activities, ActivityLaps, ActivityRecords
from .tcx import Tcx


class ActivityExporter(object):
    """Export activities as TCX files from database data."""

    def __init__(self, directory, activity_id, measurement_system, debug):
        """Return a instance of ActivityExporter ready to write a TCX file."""
        self.directory = directory
        self.activity_id = activity_id
        self.measurement_system = measurement_system
        self.debug = debug

    def process(self, db_params):
        """Process database data for an activity into a an XML tree in TCX format."""
        garmin_act_db = ActivitiesDb(db_params, self.debug - 1)
        with garmin_act_db.managed_session() as garmin_act_db_session:
            activity = Activities.s_get(garmin_act_db_session, self.activity_id)
            self.tcx = Tcx()
            self.tcx.create(activity.sport, activity.start_time)
            laps = ActivityLaps.s_get_activity(garmin_act_db_session, self.activity_id)
            records = ActivityRecords.s_get_activity(garmin_act_db_session, self.activity_id)
            for lap in laps:
                distance = Distance.from_meters_or_feet(lap.distance, self.measurement_system)
                track = self.tcx.add_lap(lap.start_time, lap.stop_time, distance, lap.calories)
                for record in records:
                    if record.timestamp >= lap.start_time and record.timestamp <= lap.stop_time:
                        alititude = Distance.from_meters_or_feet(record.altitude, self.measurement_system)
                        speed = Speed.from_kph_or_mph(record.speed, self.measurement_system)
                        self.tcx.add_point(track, record.timestamp, record.position, alititude, record.hr, speed)
        gdb = GarminDb(db_params)
        with gdb.managed_session() as garmin_db_session:
            file = File.s_get(garmin_db_session, self.activity_id)
            device = Device.s_get(garmin_db_session, file.serial_number)
            self.tcx.add_creator(device.product, file.serial_number)

    def write(self, filename):
        """Write the TCX file to disk."""
        full_path = self.directory + os.path.sep + filename
        self.tcx.write(full_path)
        return full_path



"""Objects representing a database and database objects for storing health data from a Microsoft Health."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import datetime
import logging
from sqlalchemy import Column, Integer, Date, DateTime, Float

import fitfile.conversions as conversions
import idbutils


logger = logging.getLogger(__name__)

MSHealthDb = idbutils.DB.create('mshealth', 2, "Database for storing health data from Microsoft Health.")
Attributes = idbutils.DbObject.create('attributes', MSHealthDb, 1, base=idbutils.KeyValueObject, doc="key-value data from a Microsoft Health device.")


class DaysSummary(MSHealthDb.Base, idbutils.DbObject):
    """A table that holds summarized information about a day with one row per day."""

    __tablename__ = 'days_summary'

    db = MSHealthDb
    table_version = 1

    day = Column(Date, primary_key=True)
    calories = Column(Integer)
    steps = Column(Integer)
    floors = Column(Integer)
    hr_max = Column(Integer)
    hr_min = Column(Integer)
    hr_avg = Column(Integer)
    active_hours = Column(Integer)
    activity_secs = Column(Integer)
    activity_calories = Column(Integer)
    exercise_events = Column(Integer)
    exercise_calories = Column(Integer)
    exercise_secs = Column(Integer)
    miles_moved = Column(Float)
    sleep_events = Column(Integer)
    sleep_calories = Column(Integer)
    sleep_secs = Column(Integer)
    walk_events = Column(Integer)
    walk_secs = Column(Integer)
    workout_calories = Column(Integer)
    miles_walked = Column(Float)
    run_ewvents = Column(Integer)
    run_calories = Column(Integer)
    run_secs = Column(Integer)
    miles_run = Column(Float)
    miles_golfed = Column(Integer)
    golf_calories = Column(Integer)
    golf_events = Column(Integer)
    golf_secs = Column(Integer)
    miles_biked = Column(Float)
    uv_mins = Column(Integer)
    bike_secs = Column(Integer)
    bike_calories = Column(Integer)
    bike_events = Column(Integer)
    guided_workout_events = Column(Integer)
    guided_workout_calories = Column(Integer)
    guided_workout_secs = Column(Integer)

    @classmethod
    def get_hr_stats(cls, db, start_ts, end_ts):
        """Return a dictionary of aggregate heart rate statistics for the given time period."""
        return {
            'hr_avg' : cls.get_col_avg(db, cls.hr_avg, start_ts, end_ts, True),
            'hr_min' : cls.get_col_min(db, cls.hr_min, start_ts, end_ts, True),
            'hr_max' : cls.get_col_max(db, cls.hr_max, start_ts, end_ts),
        }

    @classmethod
    def get_activity_mins_stats(cls, db, func, start_ts, end_ts):
        """Return a dictionary of aggregate activity mins statistics for the given time period."""
        active_hours = func(db, cls.active_hours, start_ts, end_ts)
        if active_hours is not None:
            intensity_time = conversions.hours_to_dt_time(active_hours)
            stats = {
                'intensity_time' : intensity_time,
                'moderate_activity_time' : intensity_time,
                # 'vigorous_activity_time' : None,      Don't write where we have no data, may overwrite good data
            }
        else:
            stats = {}
        return stats

    @classmethod
    def get_floors_stats(cls, db, func, start_ts, end_ts):
        """Return a dictionary of aggregate floors statistics for the given time period."""
        return {'floors' : func(db, cls.floors, start_ts, end_ts)}

    @classmethod
    def get_steps_stats(cls, db, func, start_ts, end_ts):
        """Return a dictionary of aggregate steps statistics for the given time period."""
        return {'steps' : func(db, cls.steps, start_ts, end_ts)}

    @classmethod
    def get_sleep_stats(cls, db, start_ts, end_ts):
        """Return a dictionary of aggregate sleep statistics for the given time period."""
        return {
            'sleep_avg' : conversions.secs_to_dt_time(cls.get_col_avg(db, cls.sleep_secs, start_ts, end_ts, True)),
            'sleep_min' : conversions.secs_to_dt_time(cls.get_col_min(db, cls.sleep_secs, start_ts, end_ts, True)),
            'sleep_max' : conversions.secs_to_dt_time(cls.get_col_max(db, cls.sleep_secs, start_ts, end_ts)),
        }

    @classmethod
    def get_calories_stats(cls, db, start_ts, end_ts):
        """Return a dictionary of aggregate calorie statistics for the given time period."""
        calories_avg = cls.get_col_avg(db, cls.calories, start_ts, end_ts)
        calories_active_avg = cls.get_col_avg(db, cls.activity_calories, start_ts, end_ts)
        if calories_active_avg is not None:
            calories_bmr_avg = calories_avg - calories_active_avg
        else:
            calories_bmr_avg = calories_avg
        return {
            'calories_avg'          : calories_avg,
            'calories_bmr_avg'      : calories_bmr_avg,
            'calories_active_avg'   : calories_active_avg,
        }

    @classmethod
    def get_daily_stats(cls, db, day_ts):
        """Return a dictionary of aggregate statistics for the given day."""
        stats = cls.get_activity_mins_stats(db, cls.get_col_sum, day_ts, day_ts + datetime.timedelta(1))
        stats.update(cls.get_floors_stats(db, cls.get_col_sum, day_ts, day_ts + datetime.timedelta(1)))
        stats.update(cls.get_steps_stats(db, cls.get_col_sum, day_ts, day_ts + datetime.timedelta(1)))
        stats.update(cls.get_hr_stats(db, day_ts, day_ts + datetime.timedelta(1)))
        stats.update(cls.get_sleep_stats(db, day_ts, day_ts + datetime.timedelta(1)))
        stats.update(cls.get_calories_stats(db, day_ts, day_ts + datetime.timedelta(1)))
        stats['day'] = day_ts
        return stats

    @classmethod
    def get_weekly_stats(cls, db, first_day_ts):
        """Return a dictionary of aggregate statistics for the given week."""
        stats = cls.get_activity_mins_stats(db, cls.get_col_sum, first_day_ts, first_day_ts + datetime.timedelta(7))
        stats.update(cls.get_floors_stats(db, cls.get_col_sum, first_day_ts, first_day_ts + datetime.timedelta(7)))
        stats.update(cls.get_steps_stats(db, cls.get_col_sum, first_day_ts, first_day_ts + datetime.timedelta(7)))
        stats.update(cls.get_hr_stats(db, first_day_ts, first_day_ts + datetime.timedelta(7)))
        stats.update(cls.get_sleep_stats(db, first_day_ts, first_day_ts + datetime.timedelta(7)))
        stats.update(cls.get_calories_stats(db, first_day_ts, first_day_ts + datetime.timedelta(7)))
        stats['first_day'] = first_day_ts
        return stats

    @classmethod
    def get_monthly_stats(cls, db, first_day_ts, last_day_ts):
        """Return a dictionary of aggregate statistics for the given month."""
        stats = cls.get_activity_mins_stats(db, cls.get_col_sum, first_day_ts, last_day_ts)
        stats.update(cls.get_floors_stats(db, cls.get_col_sum, first_day_ts, last_day_ts))
        stats.update(cls.get_steps_stats(db, cls.get_col_sum, first_day_ts, last_day_ts))
        stats.update(cls.get_hr_stats(db, first_day_ts, last_day_ts))
        stats.update(cls.get_sleep_stats(db, first_day_ts, last_day_ts))
        stats.update(cls.get_calories_stats(db, first_day_ts, last_day_ts))
        stats['first_day'] = first_day_ts
        return stats

    @classmethod
    def get_yearly_stats(cls, db, year):
        """Return a dictionary of aggregate statistics for the given year."""
        first_day_ts = datetime.datetime(year, 1, 1)
        last_day_ts = first_day_ts + datetime.timedelta(365)
        stats = cls.get_activity_mins_stats(db, cls.get_col_sum, first_day_ts, last_day_ts)
        stats.update(cls.get_floors_stats(db, cls.get_col_sum, first_day_ts, last_day_ts))
        stats.update(cls.get_steps_stats(db, cls.get_col_sum, first_day_ts, last_day_ts))
        stats.update(cls.get_hr_stats(db, first_day_ts, last_day_ts))
        stats.update(cls.get_sleep_stats(db, first_day_ts, last_day_ts))
        stats.update(cls.get_calories_stats(db, first_day_ts, last_day_ts))
        stats['first_day'] = first_day_ts
        return stats


class MSVaultWeight(MSHealthDb.Base, idbutils.DbObject):
    """Class for a databse table holding weight data from Microsoft Health Vault."""

    __tablename__ = 'weight'

    db = MSHealthDb
    table_version = 1

    timestamp = Column(DateTime, primary_key=True, unique=True)
    weight = Column(Float)

    @classmethod
    def get_stats(cls, db, start_ts, end_ts):
        """Return a dictionary of aggregate statistics for the given time period."""
        stats = {
            'weight_avg' : cls.get_col_avg(db, cls.weight, start_ts, end_ts, True),
            'weight_min' : cls.get_col_min(db, cls.weight, start_ts, end_ts, True),
            'weight_max' : cls.get_col_max(db, cls.weight, start_ts, end_ts),
        }
        return stats



"""Test Garmin monitoring database data."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging
import datetime

import fitfile

from garmindb import ConfigManager, GarminMonitoringFitData, GarminSummaryData, MonitoringFitFileProcessor, PluginManager
from garmindb.garmindb import GarminDb, File, Device, DeviceInfo, DailySummary
from garmindb.garmindb import MonitoringDb, Monitoring, MonitoringInfo, MonitoringHeartRate, MonitoringIntensity, MonitoringClimb

from test_db_base import TestDBBase


root_logger = logging.getLogger()
handler = logging.FileHandler('monitoring_db.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestMonitoringDB(TestDBBase, unittest.TestCase):
    """Class for testing Garmin monitoring database data."""

    @classmethod
    def setUpClass(cls):
        db_params = ConfigManager.get_db_params()
        cls.plugin_manager = PluginManager(ConfigManager.get_or_create_plugins_dir(), db_params)
        cls.garmin_mon_db = MonitoringDb(db_params)
        table_dict = {
            'monitoring_info_table'         : MonitoringInfo,
            'monitoring_hr_table'           : MonitoringHeartRate,
            'monitoring_intensity_table'    : MonitoringIntensity,
            'monitoring_climb_table'        : MonitoringClimb,
            'monitoring_table'              : Monitoring,
        }
        super().setUpClass(cls.garmin_mon_db, table_dict)

    def test_garmin_mon_db_tables_exists(self):
        self.assertGreater(MonitoringInfo.row_count(self.db), 0)
        self.assertGreater(MonitoringHeartRate.row_count(self.db), 0)
        self.assertGreater(MonitoringIntensity.row_count(self.db), 0)
        self.assertGreater(MonitoringClimb.row_count(self.db), 0)
        self.assertGreater(Monitoring.row_count(self.db), 0)

    def test_garmin_mon_db_steps_bounds(self):
        min = Monitoring.get_col_min(self.db, Monitoring.steps)
        self.assertGreater(min, 0)
        max = Monitoring.get_col_max(self.db, Monitoring.steps)
        self.assertGreater(max, 0)
        self.assertLess(max, 100000)

    def test_garmin_mon_db_uptodate(self):
        uptodate_tables = {
            'monitoring_hr_table'   : MonitoringHeartRate,
            'monitoring_table'      : Monitoring,
        }
        for table_name, table in uptodate_tables.items():
            latest = MonitoringHeartRate.latest_time(self.db, MonitoringHeartRate.heart_rate)
            logger.info("Latest data for %s: %s", table_name, latest)
            self.assertLess(datetime.datetime.now() - latest, datetime.timedelta(days=2))

    def fit_file_import(self, db_params):
        gfd = GarminMonitoringFitData('test_files/fit/monitoring', latest=False, measurement_system=fitfile.field_enums.DisplayMeasure.statute, debug=2)
        self.gfd_file_count = gfd.file_count()
        if gfd.file_count() > 0:
            gfd.process_files(MonitoringFitFileProcessor(db_params, self.plugin_manager))

    def test_fit_file_import(self):
        db_params = ConfigManager.get_db_params(test_db=True)
        self.profile_function('fit_mon_import', self.fit_file_import, db_params)
        test_mon_db = GarminDb(db_params)
        self.check_db_tables_exists(test_mon_db, {'device_table' : Device})
        self.check_db_tables_exists(test_mon_db, {'file_table' : File, 'device_info_table' : DeviceInfo}, self.gfd_file_count)
        table_not_none_cols_dict = {Monitoring : [Monitoring.timestamp, Monitoring.activity_type, Monitoring.duration]}
        self.check_not_none_cols(MonitoringDb(db_params), table_not_none_cols_dict)

    def test_summary_json_file_import(self):
        db_params = ConfigManager.get_db_params(test_db=True)
        gjsd = GarminSummaryData(db_params, 'test_files/json/monitoring/summary', latest=False, measurement_system=fitfile.field_enums.DisplayMeasure.statute, debug=2)
        if gjsd.file_count() > 0:
            gjsd.process()
        table_not_none_cols_dict = {
            DailySummary : [DailySummary.rhr, DailySummary.distance, DailySummary.steps, DailySummary.floors_goal]
        }
        self.check_not_none_cols(GarminDb(db_params), table_not_none_cols_dict)

    def check_day_steps(self, data):
        last_steps = {}
        last_steps_timestamp = None
        for monitoring in data:
            if monitoring.steps is not None:
                steps = monitoring.steps
                activity_type = monitoring.activity_type
                if activity_type in last_steps:
                    activity_last_steps = last_steps[activity_type]
                    self.assertGreaterEqual(steps, activity_last_steps,
                                            f'{repr(monitoring)} - steps at {monitoring.timestamp} not greater than last steps at {last_steps_timestamp}')
                last_steps[activity_type] = steps
                last_steps_timestamp = monitoring.timestamp

    def test_db_data_integrity(self):
        target_day = (datetime.datetime.now() - datetime.timedelta(days=2)).date()
        day_data = Monitoring.get_for_day(self.garmin_mon_db, Monitoring, target_day)
        self.check_day_steps(day_data)


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3

"""Script for importing into a DB and summarizing CSV formatted Microsoft Health export data."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import sys
import argparse
import logging

from garmindb import ConfigManager, format_version
from garmindb.mshealthdb import MSHealthDb, MSHealthData, MSVaultData, Analyze


logging.basicConfig(filename='mshealth.log', filemode='w', level=logging.DEBUG)
logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


def __usage(program):
    print('%s -i <inputfile>' % program)
    sys.exit()


def main(argv):
    """Import and analyze Microsoft Health data."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", help="print the program's version", action='version', version=format_version(sys.argv[0]))
    parser.add_argument("-t", "--trace", help="Turn on debug tracing", type=int, default=0)
    modes_group = parser.add_argument_group('Modes')
    modes_group.add_argument("-i", "--input_file", help="Specifiy the CSV file to import into the database")
    modes_group.add_argument("--delete_db", help="Delete MSHealth db file.", action="store_true", default=False)
    args = parser.parse_args()

    root_logger = logging.getLogger()
    if args.trace:
        root_logger.setLevel(logging.DEBUG)
    else:
        root_logger.setLevel(logging.INFO)

    db_params = ConfigManager.get_db_params()

    if args.delete_db:
        MSHealthDb.delete_db(db_params)
        sys.exit()

    mshealth_dir = ConfigManager.get_or_create_mshealth_dir()
    metric = ConfigManager.get_metric()

    msd = MSHealthData(args.input_file, mshealth_dir, db_params, metric, args.trace)
    if msd.file_count() > 0:
        msd.process_files()

    mshv = MSVaultData(args.input_file, mshealth_dir, db_params, metric, args.trace)
    if mshv.file_count() > 0:
        mshv.process_files()

    analyze = Analyze(db_params)
    analyze.get_years()
    analyze.summary()


if __name__ == "__main__":
    main(sys.argv[1:])
